# coding=utf-8
""" Python AST visitor that converts an expression to a RACO equivalent """

import sys
import ast

from raco.expression import NamedAttributeRef, UnnamedAttributeRef, \
    StringLiteral, NumericLiteral, BooleanLiteral, \
    EQ, NEQ, LT, LTEQ, GT, GTEQ, AND, OR, NOT, \
    PLUS, MINUS, DIVIDE, IDIVIDE, MOD, TIMES, NEG, CAST, PYUDF
from raco.types import STRING_TYPE, LONG_TYPE, DOUBLE_TYPE, BOOLEAN_TYPE
from raco.expression.function import WORKERID, RANDOM, \
    ABS, CEIL, COS, FLOOR, LOG, SIN, SQRT, TAN, MD5, LEN, POW, \
    LESSER, GREATER, SUBSTR
from raco.python.exceptions import PythonUnrecognizedTokenException, \
    PythonOutOfRangeException, PythonUnsupportedOperationException, \
    PythonArgumentException

comparator_map = {
    ast.Eq: EQ,
    ast.NotEq: NEQ,
    ast.Lt: LT,
    ast.LtE: LTEQ,
    ast.Gt: GT,
    ast.GtE: GTEQ
}

propositional_map = {
    ast.And: AND,
    ast.Or: OR
}

unary_operators = {
    ast.Not: NOT,
    ast.USub: NEG
}

binary_operators = {
    ast.Add: PLUS,
    ast.Sub: MINUS,
    ast.Div: DIVIDE,
    ast.FloorDiv: IDIVIDE,
    ast.Mod: MOD,
    ast.Mult: TIMES
}

literal_map = {
    'True': BooleanLiteral(True),
    'False': BooleanLiteral(False),
}

nary_map = {
    # Arity, name: function
    (1, 'str'): lambda args: CAST(STRING_TYPE, args[0]),
    (1, 'int'): lambda args: MOD(CAST(LONG_TYPE, args[0]),
                                 NumericLiteral(sys.maxint)),
    (1, 'long'): lambda args: CAST(LONG_TYPE, args[0]),
    (1, 'float'): lambda args: CAST(DOUBLE_TYPE, args[0]),
    (1, 'bool'): lambda args: CAST(BOOLEAN_TYPE, args[0]),

    (0, 'workerid'): lambda args: WORKERID(),
    (0, 'random'): lambda args: RANDOM(),
    (1, 'fabs'): lambda args: ABS(CAST(DOUBLE_TYPE, args[0])),
    (1, 'abs'): lambda args: ABS(args[0]),
    (1, 'ceil'): lambda args: CEIL(args[0]),
    (1, 'cos'): lambda args: COS(args[0]),
    (1, 'floor'): lambda args: FLOOR(args[0]),
    (1, 'log'): lambda args: LOG(args[0]),
    (1, 'sin'): lambda args: SIN(args[0]),
    (1, 'sqrt'): lambda args: SQRT(args[0]),
    (1, 'tan'): lambda args: TAN(args[0]),
    (1, 'md5'): lambda args: MD5(args[0]),
    (1, 'len'): lambda args: LEN(args[0]),
    (2, 'pow'): lambda args: POW(args[0], args[1]),

    (2, 'min'): lambda args: LESSER(args[0], args[1]),
    (2, 'max'): lambda args: GREATER(args[0], args[1]),
}

zero = ast.Num(n=0)


class ExpressionVisitor(ast.NodeVisitor):
    """ Visitor that converts an AST to a RACO expression """

    def __init__(self, schema, udfs):
        self.schema = schema
        self.names = None
        self.udfs = udfs

    def visit_arguments(self, node):
        """ Visitor for function arguments """
        self.names = [n.id for n in node.args]

    def visit_UnaryOp(self, node):
        """ Visitor for a unary operator """
        if type(node.op) not in unary_operators:
            raise PythonUnrecognizedTokenException(node.op,
                                                   node.lineno,
                                                   node.col_offset)
        return unary_operators[type(node.op)](self.visit(node.operand))

    def visit_BinOp(self, node):
        """ Visitor for binary operations """
        if type(node.op) not in binary_operators:
            raise PythonUnrecognizedTokenException(node.op,
                                                   node.lineno,
                                                   node.col_offset)
        return binary_operators[type(node.op)](
            self.visit(node.left), self.visit(node.right))

    def visit_BoolOp(self, node):
        """ Visitor for boolean operations """
        assert (len(node.values) >= 2)
        op = propositional_map[type(node.op)]
        return reduce(lambda c, e: op(c, self.visit(e)),
                      # Fold over any other clauses
                      node.values[2:],
                      # Always at least two clauses
                      op(*map(self.visit, node.values[:2])))

    def visit_Compare(self, node):
        """ Visitor for comparison operations """
        if len(node.ops) == 1 and \
           type(node.ops[0]) in comparator_map.keys() and \
           len(node.comparators) == 1:
            left = self.visit(node.left)
            right = self.visit(node.comparators[0])
            return comparator_map[type(node.ops[0])](left, right)
        else:
            raise PythonUnrecognizedTokenException(node.ops[0],
                                                   node.lineno,
                                                   node.col_offset)

    def visit_Attribute(self, node):
        """ Visitor for dotted references """
        if not isinstance(node.value, ast.Name):
            raise PythonUnsupportedOperationException(
                'Unsupported reference',
                node.lineno, node.col_offset)

        scheme = self.schema[self.names.index(node.value.id)]

        if node.attr not in [s[0] for s in scheme]:
            raise PythonUnrecognizedTokenException(node.attr,
                                                   node.lineno,
                                                   node.col_offset)

        return NamedAttributeRef(node.attr)

    def visit_Subscript(self, node):
        """ Visitor for slices """
        if isinstance(node.value, ast.Name) and \
           isinstance(node.slice, ast.Index) and \
           node.value.id in self.names:
            return self.visit_AttributeIndex(node)
        elif isinstance(node.slice, ast.Slice):
            return self.visit_SubstringSlice(node)
        elif isinstance(node.slice, ast.Index):
            return self.visit_SubstringIndex(node)

    def visit_AttributeIndex(self, node):
        """ Visitor for subscripts over a tuple """
        if node.value.id not in self.names:
            raise PythonUnrecognizedTokenException(
                node.value.id, node.lineno, node.col_offset)

        offset = sum([len(s) for s in
                      self.schema[:self.names.index(node.value.id)]])

        if node.slice.value.n < 0 or \
           node.slice.value.n >= \
                len(self.schema[self.names.index(node.value.id)]):
            raise PythonOutOfRangeException(
                node.value.id, node.lineno, node.col_offset)

        return UnnamedAttributeRef(node.slice.value.n + offset)

    def visit_SubstringIndex(self, node):
        """ Visitor for indexing a string """
        node.slice = ast.Slice(lower=node.slice.value,
                               upper=ast.Num(node.slice.value.n + 1),
                               step=None)
        return self.visit_SubstringSlice(node)

    def visit_SubstringSlice(self, node):
        """ Visitor for slicing a string """
        if (node.slice.lower or zero).n < 0 or \
           (node.slice.upper or zero).n < 0:
            raise PythonUnsupportedOperationException(
                'RACO does not support negative indices in slices',
                node.lineno, node.col_offset)
        elif node.slice.step is not None:
            raise PythonUnsupportedOperationException(
                'RACO does not support steps in slices',
                node.lineno, node.col_offset)

        child = self.visit(node.value)
        return SUBSTR(
            [child,
             self.visit(node.slice.lower or ast.Num(0)),
             self.visit(node.slice.upper or ast.Num(2 ** 30))])

    def visit_Call(self, node):
        """ Visitor for calling built-in or UDF functions """
        name = node.func.id
        arity = len(node.args)

        if (arity, name) in nary_map:
            return self.visit_Call_builtin(node)
        elif name in (udf['name'] for udf in self.udfs):
            return self.visit_Call_UDF(node)
        else:
            raise PythonArgumentException(
                'Unrecognized function %s or invalid arguments' % name,
                node.lineno, node.col_offset)

    def visit_Call_builtin(self, node):
        """ Visitor for calling built-in functions """
        name = node.func.id
        arity = len(node.args)

        if (arity, name) not in nary_map:
            raise PythonArgumentException(
                'Unrecognized function %s or invalid arguments' % name,
                node.lineno, node.col_offset)

        return nary_map[(arity, name)](map(self.visit, node.args))

    def visit_Call_UDF(self, node):
        """ Visitor for calling UDF functions """
        name = node.func.id
        arity = len(node.args)
        udf = next((udf for udf in self.udfs
                    if udf['name'] == name), None)

        # Ignore output type when looking up UDF
        if udf is None:
            raise PythonArgumentException(
                'Unrecognized function %s or invalid arguments' % name,
                node.lineno, node.col_offset)

        output_type = udf['outputType']
        source = udf.get('source', None)
        return PYUDF(name, output_type,
                     *map(self.visit, node.args),
                     source=source)

    def visit_Str(self, node):
        """ Visitor for string literals """
        return StringLiteral(node.s)

    def visit_Num(self, node):
        """ Visitor for numeric literals """
        return NumericLiteral(node.n)

    def visit_Name(self, node):
        """ Visitor for built-in literals """
        if node.id not in literal_map:
            raise PythonUnrecognizedTokenException(node.id,
                                                   node.lineno,
                                                   node.col_offset)
        return literal_map[node.id]

    def visit_Module(self, node):
        """ Visitor for top-level modules """
        assert(len(node.body) == 1)
        return self.visit(node.body[0])

    def visit_Lambda(self, node):
        """ Visitor for lambdas """
        self.visit(node.args)
        return self.visit(node.body)

    def visit_FunctionDef(self, node):
        """ Visitor for function declaration """
        # TODO this is not a strict RACO requirement
        if len(node.body) != 1:
            raise PythonUnsupportedOperationException(
                'Functions must have exactly one statement',
                node.lineno, node.col_offset)
        elif not isinstance(node.body[0], ast.Return):
            raise PythonUnsupportedOperationException(
                'Statement in function body must be a return',
                node.lineno, node.col_offset)

        self.visit(node.args)
        return self.visit(node.body[0])

    def visit_Return(self, node):
        """ Visitor for return statements """
        return self.visit(node.value)

    def visit_Expr(self, node):
        """ Visitor for expressions """
        return self.visit(node.value)

    def generic_visit(self, node):
        """ Visitor for unsupported node types """
        raise PythonUnsupportedOperationException(
            'Unsupported node ' + str(type(node)),
            node.lineno, node.col_offset)



# coding=utf-8
""" Tests for operators """

from collections import Counter
from raco.python import convert
from raco.algebra import Select
from python_test import PythonTestCase


class TestOperators(PythonTestCase):
    def _execute(self, query, expected):
        predicate = convert(query, [self.schema])
        self.assertIsNotNone(predicate)

        expression = Select(predicate, self.scan)
        plan = self.get_query(expression)
        return self.check_result(plan, expected)

    # Binary operators

    def test_add(self):
        self._execute("""lambda t: t[0] + 1 == 7""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_subtract(self):
        self._execute("""lambda t: t[0] - 1 == 5""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_multiply(self):
        self._execute("""lambda t: t[0] * 2 == 12""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_divide(self):
        self._execute("""lambda t: t[0] / 2 == 3.5""",
                      Counter([(7, 1, 'Magdalena Balazinska', 25000)]))

    def test_integer_divide(self):
        self._execute("""lambda t: t[0] // 2 == 3""",
                      Counter([(6, 3, "Dan Suciu", 90000),
                               (7, 1, 'Magdalena Balazinska', 25000)]))

    def test_modulo(self):
        self._execute("""lambda t: t[0] % 6 == 0""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    # Unary operators

    def test_negation(self):
        self._execute("""lambda t: -t[0] == -6""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_cast_integer(self):
        self._execute("""lambda t: int(t[0] / 2) == 3""",
                      Counter([(6, 3, "Dan Suciu", 90000),
                               (7, 1, 'Magdalena Balazinska', 25000)]))

    def test_cast_long(self):
        self._execute("""lambda t: long(t[0] / 2) == 3""",
                      Counter([(6, 3, "Dan Suciu", 90000),
                               (7, 1, 'Magdalena Balazinska', 25000)]))

    def test_cast_float(self):
        self._execute("""lambda t: float(t[0] / 2) == 3.0""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_cast_string(self):
        self._execute("""lambda t: str(t[0]) == '6'""",
                      Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_cast_boolean(self):
        self._execute("""lambda t: bool(t[0]) == False""",
                      Counter([]))

    def test_substring(self):
        self._execute("""lambda t: t.name[2:4] == 'um'""",
                      Counter([(4, 2, u'Shumo Chu', 5000)]))

    def test_substring_upper(self):
        self._execute("""lambda t: t.name[10:] == 'Balazinska'""",
                      Counter([(7, 1, u'Magdalena Balazinska', 25000)]))

    def test_substring_index(self):
        self._execute("""lambda t: t.name[10] == 'B'""",
                      Counter([(7, 1, u'Magdalena Balazinska', 25000)]))

    def test_substring_lower(self):
        self._execute("""lambda t: t.name[:9] == 'Magdalena'""",
                      Counter([(7, 1, u'Magdalena Balazinska', 25000)]))



from raco.backends.sparql import SPARQLAlgebra
from raco.platform_tests import MyriaLPlatformTestHarness
import raco.compile


class SPARQLTests(object):
    # TODO: refactor MyrialPlatformTests to share code
    def check_sub_tables(self, query, name, **kwargs):
        self.check(query % self.tables, name, **kwargs)

    def test_scan(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        STORE(T1, OUTPUT);
        """, "scan")

    def test_select(self):
        self.check_sub_tables("""
        T1 = SCAN(%(T1)s);
        x = [FROM T1 WHERE a>5 EMIT a];
        STORE(x, OUTPUT);
        """, "select")

    def test_join(self):
        self.check_sub_tables("""
        T3 = SCAN(%(T3)s);
        R3 = SCAN(%(R3)s);
        out = JOIN(T3, b, R3, b);
        out2 = [FROM out WHERE $3 = $5 EMIT $0, $3];
        STORE(out2, OUTPUT);
        """, "join")


class SPARQLMyriaLTests(MyriaLPlatformTestHarness, SPARQLTests):

    def check(self, query, name):
        plan = self.get_physical_plan(query, target_alg=SPARQLAlgebra())

        sparql = raco.compile.compile(plan)

        # TODO pretty lenient tests: is it a non empty string?
        assert isinstance(sparql, ''.__class__)
        assert sparql != ''



# coding=utf-8
""" Functions to convert Python functions to RACO expressions """

import ast

from raco.python.exceptions import PythonSyntaxException, \
    PythonConvertException
from raco.python.util import visitor
from raco.python.util.decompile import get_source


def convert(source_or_ast_or_callable, schema, udfs=None):
    """
    Convert a Python function into its RACO equivalent
    :param source_or_ast_or_callable: Source string, callable, or AST node
    :param schema: List of schema for the input parameter(s)
    :param udfs: List of (name, arity) pairs of UDFs
    :return: RACO expression representing the source, callable, or AST node
    """
    if isinstance(source_or_ast_or_callable, basestring):
        try:
            return convert(ast.parse(source_or_ast_or_callable), schema, udfs)
        except SyntaxError as e:
            raise PythonSyntaxException(e.msg, e.lineno, e.offset)
    elif callable(source_or_ast_or_callable):
        return convert(get_source(source_or_ast_or_callable), schema, udfs)
    elif isinstance(source_or_ast_or_callable, ast.AST):
        return visitor.ExpressionVisitor(schema or [], udfs or []).visit(
            source_or_ast_or_callable) or None
    else:
        raise PythonConvertException(
            'Argument was not a source string, callable, or AST node')



# coding=utf-8
""" Tests for predicates and comparators """

from collections import Counter
from raco.python import convert
from raco.algebra import Select
from python_test import PythonTestCase


class TestBooleans(PythonTestCase):
    def _execute_predicate(self, query, expected):
        predicate = convert(query, [self.schema])
        self.assertIsNotNone(predicate)

        expression = Select(predicate, self.scan)
        plan = self.get_query(expression)
        return self.check_result(plan, expected)

    def test_integer_equality(self):
        self._execute_predicate("""lambda t: t[0] == 6""",
                                Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_float_equality(self):
        self._execute_predicate("""lambda t: t[0] == 6.0""",
                                Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_string_equality(self):
        self._execute_predicate("""lambda t: t[2] == 'Dan Suciu'""",
                                Counter([(6, 3, "Dan Suciu", 90000)]))

    def test_inequality(self):
        self._execute_predicate("""lambda t: t[0] != 999""",
                                self.emp_table)

    def test_less_than(self):
        self._execute_predicate("""lambda t: t[0] < 999""",
                                self.emp_table)

    def test_less_than_equal(self):
        self._execute_predicate("""lambda t: t[0] <= 7""",
                                self.emp_table)

    def test_greater_than(self):
        self._execute_predicate("""lambda t: t[0] >= 1""",
                                self.emp_table)

    def test_greater_than_equal(self):
        self._execute_predicate("""lambda t: t[0] >= 7""",
                                Counter([(7, 1, 'Magdalena Balazinska',
                                          25000)]))

    def test_conjunction(self):
        self._execute_predicate(
            """lambda t: t[0] >= 7 and t[2] != 'Magdalena Balazinska'""",
            Counter([]))

    def test_extended_conjunction(self):
        self._execute_predicate(
            """lambda t: t[0] >= 7 and t[1] != 1 and """ +
            """                        t[2] == 'Magdalena Balazinska'""",
            Counter([]))

    def test_disjunction(self):
        self._execute_predicate(
            """lambda t: t[0] >= 999 or t[2] == 'Magdalena Balazinska'""",
            Counter([(7, 1, 'Magdalena Balazinska', 25000)]))

    def test_extended_disjunction(self):
        self._execute_predicate(
            """lambda t: t[0] >= 999 or """ +
            """          t[2] == 'Magdalena Balazinska' or 1 == 1""",
            self.emp_table)

    def test_negation(self):
        self._execute_predicate("""lambda t: not (1 == 1)""",
                                Counter([]))

    def test_mixed_clauses(self):
        self._execute_predicate(
            """lambda t: t[0] >= 0 and """ +
            """          t[2] == 'Magdalena Balazinska' or 1 == 1""",
            self.emp_table)

    def test_true(self):
        self._execute_predicate("""lambda t: True""",
                                self.emp_table)

    def test_false(self):
        self._execute_predicate("""lambda t: False""",
                                Counter([]))



# import all the expressions and algebras

from raco.backends.myria import *
from raco.backends.cpp import *
from raco.backends.cpp.cppcommon import *
from raco.backends.radish import *
from raco.algebra import *
from raco.scheme import *
from raco.expression.expression import *
from raco.expression.aggregate import *
from raco.types import *
from raco.relation_key import *
from raco.expression.boolean import *


import logging
logging.basicConfig()
_LOG = logging.getLogger(name=__name__)


def plan_from_repr(repr_string):
    _LOG.warning("Relying on eval! "
                 "This module should only be used in "
                 "trusted development situations\n")
    return eval(repr_string)



# -*- coding: UTF-8 -*-
import raco.myrial.myrial_test as myrial_test
from raco.fake_data import FakeData


class TestSamplingOperations(myrial_test.MyrialTestCase, FakeData):
    def setUp(self):
        super(TestSamplingOperations, self).setUp()

        self.db.ingest(TestSamplingOperations.emp_key,
                       TestSamplingOperations.emp_table,
                       TestSamplingOperations.emp_schema)

    def run_samplescan(self, sample_size, sample_type, is_pct=False):
        pct = '%' if is_pct else ''
        query = """
        emp = SAMPLESCAN({rel_key}, {size}{pct}, {type});
        STORE(emp, OUTPUT);
        """.format(rel_key=self.emp_key, size=sample_size, pct=pct,
                   type=sample_type)

        res = self.execute_query(query)
        if is_pct:
            expected_len = int(round(len(res) * (sample_size / 100.0)))
        else:
            expected_len = sample_size
        self.assertEquals(len(res), expected_len)

    def test_samplescan__wr_zero(self):
        self.run_samplescan(0, 'WR')

    def test_samplescan__wr_zero_pct(self):
        self.run_samplescan(0, 'WR', True)

    def test_samplescan__wor_zero(self):
        self.run_samplescan(0, 'WoR')

    def test_samplescan__wor_zero_pct(self):
        self.run_samplescan(0, 'WoR', True)

    def test_samplescan__wr_one(self):
        self.run_samplescan(1, 'WR')

    def test_samplescan__wor_one(self):
        self.run_samplescan(1, 'WoR')

    def test_samplescan__wr_all(self):
        self.run_samplescan(len(self.emp_table), 'WR')

    def test_samplescan__wr_100_pct(self):
        self.run_samplescan(100, 'WR', True)

    def test_samplescan__wor_all(self):
        self.run_samplescan(len(self.emp_table), 'WoR')

    def test_samplescan__wor_100_pct(self):
        self.run_samplescan(100, 'WoR', True)



"""
Functions (unary and binary) for use in Raco.
"""

import math
import md5
import random

from raco.expression.udf import Function

from .expression import (ZeroaryOperator, UnaryOperator, BinaryOperator,
                         NaryOperator, types, check_is_numeric, check_type,
                         TypeSafetyViolation, NamedAttributeRef)


class UnaryFunction(UnaryOperator):

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.input)

    def __repr__(self):
        return "{op}({inp!r})".format(op=self.opname(), inp=self.input)


class BinaryFunction(BinaryOperator):

    def __str__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)

    def __repr__(self):
        return "{op}({l!r}, {r!r})".format(op=self.opname(), l=self.left,
                                           r=self.right)


class CustomBinaryFunction(BinaryFunction):

    def __init__(self, name, typ, left, right):
        self.name = name
        self.typ = typ
        super(CustomBinaryFunction, self).__init__(left, right)

    def __repr__(self):
        return "{op}({n!r}, {t!r}, {l!r}, {r!r})".format(op=self.opname(),
                                                         n=self.name,
                                                         t=self.typ,
                                                         l=self.left,
                                                         r=self.right)

    def typeof(self, scheme, state_scheme):
        return self.typ

    literals = []

    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError("Not intended for evaluation")


class ZeroaryFunction(ZeroaryOperator):

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __repr__(self):
        return "{op}()".format(op=self.opname())


class CustomZeroaryFunction(ZeroaryFunction):

    def __init__(self, name, typ):
        self.name = name
        self.typ = typ
        super(CustomZeroaryFunction, self).__init__()

    def __str__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.name, self.typ)

    def __repr__(self):
        return "{op}({n!r}, {t!r})".format(op=self.opname(),
                                           n=self.name,
                                           t=self.typ)

    def typeof(self, scheme, state_scheme):
        return self.typ

    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError("Not intended for evaluation")


class NaryFunction(NaryOperator):

    def __str__(self):
        return "%s(%s)" % \
            (self.__class__.__name__,
             ",".join([str(op) for op in self.operands]))

    def __repr__(self):
        return "{op}({ch!r})".format(op=self.opname(), ch=self.operands)


class WORKERID(ZeroaryOperator):

    def __str__(self):
        return "%s" % self.__class__.__name__

    def __repr__(self):
        return "{op}()".format(op=self.opname())

    def evaluate(self, _tuple, scheme, state=None):
        return 0

    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE


class RANDOM(ZeroaryOperator):

    def __str__(self):
        return "%s" % self.__class__.__name__

    def __repr__(self):
        return "{op}()".format(op=self.opname())

    def evaluate(self, _tuple, scheme, state=None):
        return random.random()

    def typeof(self, scheme, state_scheme):
        return types.DOUBLE_TYPE


class UnaryDateToNumFunction(UnaryFunction):
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        if self.input.typeof(scheme, state_scheme) != types.STRING_TYPE:
            raise TypeSafetyViolation("Dates can only be strings")
        return types.LONG_TYPE


class YEAR(UnaryDateToNumFunction):
    pass


class MONTH(UnaryDateToNumFunction):
    pass


class DAY(UnaryDateToNumFunction):
    pass


class UnaryDoubleFunction(UnaryFunction):

    """A unary function that returns a double."""

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return types.DOUBLE_TYPE


class UnaryTypePreservingFunction(UnaryFunction):

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return input_type


class VariadicFunction(object):
    def __init__(self, ftype, name, typ, **kwargs):
        self.ftype = ftype
        self.name = name
        self.typ = typ
        self.kwargs = kwargs

    def bind(self, *args):
        return Function(
            ['arg{}'.format(i) for i in xrange(len(args))],
            self.ftype(self.name,
                       self.typ,
                       *[NamedAttributeRef('arg{}'.format(i))
                         for i in xrange(len(args))],
                       **self.kwargs))


class ABS(UnaryTypePreservingFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return abs(self.input.evaluate(_tuple, scheme, state))


class CEIL(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.ceil(self.input.evaluate(_tuple, scheme, state))


class COS(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.cos(self.input.evaluate(_tuple, scheme, state))


class FLOOR(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.floor(self.input.evaluate(_tuple, scheme, state))


class LOG(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.log(self.input.evaluate(_tuple, scheme, state))


class SIN(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.sin(self.input.evaluate(_tuple, scheme, state))


class SQRT(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.sqrt(self.input.evaluate(_tuple, scheme, state))


class TAN(UnaryDoubleFunction):

    def evaluate(self, _tuple, scheme, state=None):
        return math.tan(self.input.evaluate(_tuple, scheme, state))


class MD5(UnaryFunction):

    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE

    def evaluate(self, _tuple, scheme, state=None):
        """Preserve 64 bits of the md5 hash function."""
        m = md5.new()
        m.update(str(self.input.evaluate(_tuple, scheme, state)))
        return int(m.hexdigest(), 16) >> 64


class POW(BinaryFunction):
    literals = ['POW']

    def evaluate(self, _tuple, scheme, state=None):
        return pow(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        check_is_numeric(lt)
        rt = self.right.typeof(scheme, state_scheme)
        check_is_numeric(rt)

        return types.DOUBLE_TYPE


class CompareFunction(BinaryFunction):

    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        rt = self.right.typeof(scheme, state_scheme)
        if lt != rt:
            raise TypeSafetyViolation("Can't compare %s with %s" % (
                lt, rt))
        return lt


class LESSER(CompareFunction):
    literals = ['LESSER']

    def evaluate(self, _tuple, scheme, state=None):
        return min(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))


class GREATER(CompareFunction):
    literals = ['GREATER']

    def evaluate(self, _tuple, scheme, state=None):
        return max(self.left.evaluate(_tuple, scheme, state),
                   self.right.evaluate(_tuple, scheme, state))


class SUBSTR(NaryFunction):
    literals = ["SUBSTR"]

    def evaluate(self, _tuple, scheme, state=None):
        inputStr = self.operands[0].evaluate(_tuple, scheme, state)
        beginIdx = self.operands[1].evaluate(_tuple, scheme, state)
        endIdx = self.operands[2].evaluate(_tuple, scheme, state)
        return inputStr[beginIdx:endIdx]

    def typeof(self, scheme, state_scheme):
        check_type(self.operands[0].typeof(scheme, state_scheme), types.STRING_TYPE)  # noqa
        check_type(self.operands[1].typeof(scheme, state_scheme), types.LONG_TYPE)  # noqa
        check_type(self.operands[2].typeof(scheme, state_scheme), types.LONG_TYPE)  # noqa

        return types.STRING_TYPE


class LEN(UnaryFunction):
    literals = ["LEN"]

    def evaluate(self, _tuple, scheme, state=None):
        return len(self.input.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        if input_type != types.STRING_TYPE:
            raise TypeSafetyViolation("Must be a string for %s" % (
                self.__class__,))
        return types.LONG_TYPE


class PYUDF(NaryFunction):

    literals = []

    def __init__(self, name, typ, *args, **kwargs):
        super(PYUDF, self).__init__(tuple(args))
        self.name = name
        self.source = kwargs.get('source', None)
        self.func = eval(self.source) if self.source else None
        self.typ = typ

    def __str__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__,
                                   self.name,
                                   map(str, self.arguments),
                                   self.typ)

    def __repr__(self):
        return "{op}({n!r},{t!r},*{a!r}, source={s!r})".format(
            op=self.opname(),
            n=self.name,
            a=self.arguments,
            t=self.typ,
            s=self.source)

    @property
    def arguments(self):
        return self.operands

    @arguments.setter
    def arguments(self, value):
        self.operands = value

    def apply(self, f):
        super(PYUDF, self).apply(f)
        self.arguments = tuple(self.arguments)

    def typeof(self, scheme, state_scheme):
        return self.typ

    def set_typ(self, typ):
        self.typ = typ

    def evaluate(self, _tuple, scheme, state=None):
        if self.func:
            return self.func(*map(lambda a: a.evaluate(_tuple, scheme, state),
                                  self.arguments))
        else:
            raise NotImplementedError()


class SPLIT(BinaryFunction):
    literals = ["SPLIT"]

    # TODO: implement once fakedb can handle flatmap functions
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        check_type(lt, types.STRING_TYPE)
        rt = self.right.typeof(scheme, state_scheme)
        check_type(rt, types.STRING_TYPE)
        return types.STRING_TYPE


class SEQUENCE(UnaryFunction):
    literals = ["SEQUENCE"]

    # TODO: implement once fakedb can handle flatmap functions
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        if input_type != types.LONG_TYPE:
            raise TypeSafetyViolation("Must be an integer for %s" % (
                self.__class__,))
        return types.LONG_TYPE


class NGRAM(BinaryFunction):
    literals = ["NGRAM"]

    # TODO: implement once fakedb can handle flatmap functions
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        check_type(lt, types.STRING_TYPE)
        rt = self.right.typeof(scheme, state_scheme)
        check_type(rt, types.LONG_TYPE)
        return types.STRING_TYPE


class BITSET(UnaryFunction):
    literals = ["BITSET"]

    # TODO: implement once fakedb can handle flatmap functions
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        if input_type != types.BLOB_TYPE:
            raise TypeSafetyViolation("Must be a BLOB for %s" % (
                self.__class__,))
        return types.BOOLEAN_TYPE




# TODO: To be refactored into parallel shared memory lang,
# where you plugin in the parallel shared memory language specific codegen

import logging

import raco.scheme as scheme
from raco import algebra
from raco import expression
from raco import rules
from raco.algebra import gensym
from raco.backends import Algebra
from raco.backends.cpp import cppcommon
from raco.backends.cpp.cppcommon import StagedTupleRef, CBaseLanguage
from raco.expression import aggregate
from raco.pipelines import Pipelined
from raco.utility import emitlist

_LOG = logging.getLogger(__name__)

import itertools


class _ARRAY_REPRESENTATION:
    GLOBAL_ARRAY = 'global_array'
    SYMMETRIC_ARRAY = 'symmetric_array'


def define_cl_arg(type, name, default_value, description):
    return GrappaLanguage.cgenv().get_template(
        'define_cl_arg.cpp').render(locals())


class GrappaStagedTupleRef(StagedTupleRef):

    def __init__(self, relsym, scheme, aligned=False):
        self.aligned = aligned
        super(GrappaStagedTupleRef, self).__init__(relsym, scheme)

    def __afterDefinitionCode__(self, numfields, fieldtypes):
        # Grappa requires structures to be block aligned if they will be
        # iterated over with localizing forall
        if self.aligned:
            return "GRAPPA_BLOCK_ALIGNED"
        else:
            return ""


class GrappaLanguage(CBaseLanguage):
    _template_path = 'radish/grappa_templates'
    _cgenv = CBaseLanguage.__get_env_for_template_libraries__(_template_path)

    @classmethod
    def on_all(cls, code):
        return """on_all_cores([=] {{
        {code}
        }});
        """.format(code=code)

    @classmethod
    def cgenv(cls):
        return cls._cgenv

    @classmethod
    def base_template(cls):
        return cls.cgenv().get_template('base_query.cpp')

    @staticmethod
    def log(txt):
        return """LOG(INFO) << "%s";\n""" % txt

    @staticmethod
    def log_unquoted(code, level=0):
        if level == 0:
            log_str = "LOG(INFO)"
        else:
            log_str = "VLOG(%s)" % (level)

        return """%(log_str)s << %(code)s;\n""" % locals()

    @classmethod
    def compile_stringliteral(cls, st):
        if cls._external_indexing:
            st = cls.c_stringify(st)
            sid = cls.newstringident()
            decl = """int64_t %s;""" % (sid)
            lookup_init = GrappaLanguage.cgenv().get_template(
                'string_index_lookup.cpp').render(locals())
            build_init = """
            string_index = build_string_index("sp2bench.index");
            """

            return """(%s)""" % sid, [decl], [build_init, lookup_init]
            # raise ValueError("String Literals not supported in
            # C language: %s" % s)
        else:
            return super(GrappaLanguage, cls).compile_stringliteral(st)

    @staticmethod
    def group_wrap(ident, grpcode, attrs):
        timing_template = GrappaLanguage.cgenv().get_template(
            'grappa_group_timing.cpp')
        inner_code = grpcode

        timer_metric = None
        if attrs['type'] == 'in_memory':
            timer_metric = "in_memory_runtime"
            # only trace in_memory
            tracing_on = "Grappa::Metrics::start_tracing();"
            tracing_off = "Grappa::Metrics::stop_tracing();"
        elif attrs['type'] == 'scan':
            timer_metric = "saved_scan_runtime"
            tracing_on = ""
            tracing_off = ""

        code = emitlist(["Grappa::Metrics::reset();",
                         timing_template.render(locals())])

        return code

    @staticmethod
    def iterators_wrap(code, attrs):
        code = GrappaLanguage.on_all(code)
        code += """
        iterate(&{fragment_symbol}, &{global_syncname});
        """.format(**attrs)
        return code

    @staticmethod
    def pipeline_wrap(ident, plcode, attrs):

        def apply_wrappers(code, wrappers):
            """
            Wraps the code successively with wrappers.
            First wrapper is innermost

            @param code the initial code to wrap
            @param wrappers tuple of format (template, bindings).
            The template must include {{inner_code}}
            """
            current_result = code
            for template, bindings in wrappers:
                allbindings = bindings.copy()
                allbindings.update({'inner_code': current_result})
                current_result = template.render(allbindings)

            return current_result

        wrappers = []

        dependences = attrs.get('dependences', set())
        assert isinstance(dependences, set)
        _LOG.debug("pipeline %s dependences %s", ident, dependences)

        name = "pipeline_{ident}".format(ident=ident)

        syncname = attrs.get('sync')
        if syncname:
            dependence_captures = emitlist(
                ["&{dep},".format(dep=d) for d in dependences])
            sync_template = GrappaLanguage.cgenv().get_template('spawn.cpp')
            wrappers.append((sync_template, locals()))

        return apply_wrappers(plcode, wrappers)


class GrappaOperator (Pipelined, algebra.Operator):
    _language = GrappaLanguage

    @classmethod
    def new_tuple_ref(cls, sym, scheme):
        return GrappaStagedTupleRef(sym, scheme)

    @classmethod
    def language(cls):
        return cls._language

    def postorder_traversal(self, func):
        return self.postorder(func)


from raco.algebra import UnaryOperator


def create_pipeline_synchronization(state):
    """
    The pipeline_synchronization will sync tasks
    within a single pipeline. Adds this new object to
    the compiler state.
    """
    global_syncname = gensym()

    # true = tracked by gce user metrics

    global_sync_decl = GrappaLanguage.cgenv().get_template(
        'sync_declaration.cpp').render(locals())

    gce_metric_template = GrappaLanguage.cgenv().get_template(
        'gce_app_metric.cpp')

    pipeline_id = state.getCurrentPipelineId()
    gce_metric_def = gce_metric_template.render(locals())

    state.addDeclarations([global_sync_decl, gce_metric_def])

    state.setPipelineProperty('global_syncname', global_syncname)
    return global_syncname


class GrappaSingletonRelation(algebra.SingletonRelation, GrappaOperator):

    def produce(self, state):
        stagedTuple = self.new_tuple_ref(gensym(), self.scheme())

        state.addDeclarations([stagedTuple.generateDefinition(),
                               "{type} {name};".format(
                                   type=stagedTuple.getTupleTypename(),
                                   name=stagedTuple.name
        )])
        inner_code = self.parent().consume(stagedTuple, self, state)

        get_pipeline_task_name(state)
        state.setPipelineProperty('type', 'in_memory')
        state.setPipelineProperty('source', self.__class__)
        state.addPipeline(inner_code)

    def consume(self, t, src, state):
        assert False, "This is a source; it does not consume"

    def shortStr(self):
        return "GrappaSingletonRelation"


# TODO: make filescan pipeline turn into
# TODO sequence(storetemp(filescan)), scantemp).
# TODO The MemoryScan should just already have the name
class GrappaMemoryScan(algebra.UnaryOperator, GrappaOperator):

    def __init__(self, inp,
                 representation=_ARRAY_REPRESENTATION.GLOBAL_ARRAY,
                 symbol_name=None,
                 analyzed_num_tuples=10000):
        self.array_representation = representation
        self.symbol_name = symbol_name
        self.analyzed_num_tuples = analyzed_num_tuples
        super(GrappaMemoryScan, self).__init__(inp)

    def num_tuples(self):
        return self.analyzed_num_tuples

    def produce(self, state):
        self.input.produce(state)

    # TODO: when have pipeline tree representation,
    # will have a consumeMaterialized() method instead;
    # for now we reuse the tuple-based consume
    def consume(self, inputsym, src, state):
        # generate the materialization from file into memory

        # scan from index
        # memory_scan_template = """forall_localized( %(inputsym)s_index->vs, \
        # %(inputsym)s_index->nv, [](int64_t ai, Vertex& a) {
        #      forall_here_async<&impl::local_gce>( 0, a.nadj, \
        # [=](int64_t start, int64_t iters) {
        #      for (int64_t i=start; i<start+iters; i++) {
        #        auto %(tuple_name)s = a.local_adj[i];
        #
        #          %(inner_plan_compiled)s
        #       } // end scan over %(inputsym)s (for)
        #       }); // end scan over %(inputsym)s (forall_here_async)
        #       }); // end scan over %(inputsym)s (forall_localized)
        #       """

        global_syncname = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        # check if the input symbol was from a temp relation
        if isinstance(inputsym, TempRelationSymbol):
            inputsym_raw = inputsym.symbol()
            inputsym = inputsym.symbol()
            readfrom = "{inputsym}.read()".format(inputsym=inputsym)
        else:
            inputsym_raw = inputsym
            readfrom = "{inputsym}.data".format(inputsym=inputsym)

        # necessary to look this up rather than create a new
        # one because we need to read from the correct type provided
        # by the File/Temp relation input
        stagedTuple = state.lookupTupleDef(inputsym_raw)

        # get template for the scan/iteration
        memory_scan_template_name = {
            _ARRAY_REPRESENTATION.GLOBAL_ARRAY:
            'global_array_memory_scan.cpp',
            _ARRAY_REPRESENTATION.SYMMETRIC_ARRAY:
            'symmetric_array_memory_scan.cpp'
        }[self.array_representation]
        memory_scan_template = self.language().cgenv().get_template(
            memory_scan_template_name)

        tuple_type = stagedTuple.getTupleTypename()
        tuple_name = stagedTuple.name

        inner_code = self.parent().consume(stagedTuple, self, state)

        code = memory_scan_template.render(locals())
        state.setPipelineProperty('type', 'in_memory')
        state.setPipelineProperty('source', self.__class__)
        state.addPipeline(code)
        return None

    def shortStr(self):
        return "{op}({rep}, {debug_info}) "\
            .format(op=self.opname(),
                    rep=self.array_representation,
                    debug_info=self.symbol_name)

    def partitioning(self):
        raise NotImplementedError()

    def __eq__(self, other):
        """
        See important __eq__ notes below
        @see FileScan.__eq__
        """
        return UnaryOperator.__eq__(self, other)

    def __repr__(self):
        return "{op}({inp!r}, {rep!r})".format(op=self.opname(),
                                               inp=self.input,
                                               rep=self.array_representation)


class CKeyUtils:

    @staticmethod
    def _aggregate_val(tuple, cols):
        return "std::make_tuple({0})".format(
            ','.join([tuple.get_code(p) for p in cols]))

    @staticmethod
    def _aggregate_type(language, sch, cols):
        return "std::tuple<{0}>".format(
            ','.join([language.typename(
                expression.UnnamedAttributeRef(c).typeof(sch, None))
                for c in cols]))


class GrappaJoin(algebra.Join, GrappaOperator):
    pass


class GrappaSymmetricHashJoin(GrappaJoin, GrappaOperator):
    _i = 0

    @classmethod
    def __genBaseName__(cls):
        name = "%03d" % cls._i
        cls._i += 1
        return name

    def __getHashName__(self):
        name = "%s_dhash_%s" % (self.__class__.__name__, self.symBase)
        return name

    def __init__(self, *args):
        super(GrappaSymmetricHashJoin, self).__init__(*args)
        self._cgenv = cppcommon.prepend_template_relpath(
            self.language().cgenv(),
            '{0}/symmetrichashjoin'.format(GrappaLanguage._template_path))

    def _force_right_to_left_sync(self):
        """Returns whether or not to create a dependence from right
        pipeline to left pipeline. Default symmetric hash join plan allows
        the two sides to be concurrent"""
        return False

    def produce(self, state):
        self.symBase = self.__genBaseName__()

        init_template = self._cgenv.get_template('hash_init.cpp')

        declr_template = self._cgenv.get_template('hash_declaration.cpp')

        my_sch = self.scheme()
        left_sch = self.left.scheme()
        right_sch = self.right.scheme()

        self.leftcols, self.rightcols = \
            algebra.convertcondition(self.condition,
                                     len(left_sch),
                                     left_sch + right_sch)

        # declaration of hash map
        self._hashname = self.__getHashName__()
        keytype = CKeyUtils._aggregate_type(
            self.language(),
            right_sch,
            self.rightcols)
        hashname = self._hashname
        self.leftTypeRef = state.createUnresolvedSymbol()
        left_in_tuple_type = self.leftTypeRef.getPlaceholder()
        self.rightTypeRef = state.createUnresolvedSymbol()
        right_in_tuple_type = self.rightTypeRef.getPlaceholder()
        hashdeclr = declr_template.render(locals())

        state.addDeclarationsUnresolved([hashdeclr])

        self.outTuple = GrappaStagedTupleRef(gensym(), my_sch)
        out_tuple_type_def = self.outTuple.generateDefinition()
        state.addDeclarations([out_tuple_type_def])

        self.right.childtag = "right"
        self.initializer = init_template.render(locals())
        state.addInitializers([self.initializer])
        self.right.produce(state)

        self.left.childtag = "left"
        self.left.produce(state)

    def consume(self, t, src, state):
        access_template = self._cgenv.get_template('hash_insert_lookup.cpp')

        hashname = self._hashname
        keyname = t.name
        side = src.childtag

        outTuple = self.outTuple
        out_tuple_type = self.outTuple.getTupleTypename()
        out_tuple_name = self.outTuple.name

        global_syncname = state.getPipelineProperty('global_syncname')

        # note that this recycles both sides of the hash join structure
        state.recordCodeWhenInLoop(
            self.language().comment("recycle") +
            "{hashname}.clear();\n".format(
                hashname=hashname))

        if src.childtag == "right":
            left_sch = self.left.scheme()

            # save for later
            self.right_in_tuple_type = t.getTupleTypename()
            state.resolveSymbol(self.rightTypeRef, self.right_in_tuple_type)

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            keyval = CKeyUtils._aggregate_val(t, self.rightcols)

            other_tuple_type = self.leftTypeRef.getPlaceholder()
            left_type = other_tuple_type
            right_type = self.right_in_tuple_type
            left_name = gensym()
            right_name = keyname
            self.right_name = right_name
            valname = left_name

            append_func_name, combine_function_def = \
                GrappaStagedTupleRef.get_append(
                    out_tuple_type,
                    left_type, len(left_sch),
                    right_type, len(t.scheme))

            # Need to add later because requires left tuple type decl.
            # Needs to be a list because could be multiple right sides
            # occurences
            if not hasattr(self, 'right_combine_decls'):
                self.right_combine_decls = []
            self.right_combine_decls.append(combine_function_def)

            if self._force_right_to_left_sync():
                self.right_syncname = get_pipeline_task_name(state)

            code = access_template.render(locals())
            return code

        if src.childtag == "left":
            if self._force_right_to_left_sync():
                state.addToPipelinePropertySet(
                    'dependences',
                    self.right_syncname)

            right_in_tuple_type = self.right_in_tuple_type
            left_in_tuple_type = t.getTupleTypename()
            state.resolveSymbol(self.leftTypeRef, left_in_tuple_type)

            keyval = CKeyUtils._aggregate_val(t, self.leftcols)

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            left_type = left_in_tuple_type
            right_type = self.right_in_tuple_type
            other_tuple_type = self.right_in_tuple_type
            left_name = keyname
            right_name = gensym()
            valname = right_name

            append_func_name, combine_function_def = \
                GrappaStagedTupleRef.get_append(
                    out_tuple_type,
                    left_type, len(t.scheme),
                    right_type, len(self.right.scheme()))

            state.addDeclarations(
                self.right_combine_decls +
                [combine_function_def])

            code = access_template.render(locals())
            return code

        assert False, "src not equal to left or right"


class GrappaOverSynchronizedSymmetricHashJoin(GrappaSymmetricHashJoin):

    def _force_right_to_left_sync(self):
        return True


class GrappaShuffleHashJoin(algebra.Join, GrappaOperator):
    # this class is out-of-date, but can be fixed
    pass
#    _i = 0
#
#    @classmethod
#    def __genBaseName__(cls):
#        name = "%03d" % cls._i
#        cls._i += 1
#        return name
#
#    def __getHashName__(self):
#        name = "hashjoin_reducer_%s" % self.symBase
#        return name
#
#    def __init__(self, *args):
#        super(GrappaShuffleHashJoin, self).__init__(*args)
#        self._cgenv = cppcommon.prepend_template_relpath(
#            self.language().cgenv(),
#            '{0}/shufflehashjoin'.format(GrappaLanguage._template_path))
#
#    def produce(self, state):
#        left_sch = self.left.scheme()
#
#        self.syncnames = []
#        self.symBase = self.__genBaseName__()
#
#        self.right.childtag = "right"
#        self.rightTupleTypeRef = None  # may remain None if CSE succeeds
#        self.leftTupleTypeRef = None  # may remain None if CSE succeeds
#
#        # find the attribute that corresponds to the right child
#        self.rightCondIsRightAttr = \
#            self.condition.right.position >= len(left_sch)
#        self.leftCondIsRightAttr = \
#            self.condition.left.position >= len(left_sch)
#        assert self.rightCondIsRightAttr ^ self.leftCondIsRightAttr
#
#        # find right key position
#        if self.rightCondIsRightAttr:
#            self.right_keypos = self.condition.right.position \
#                - len(left_sch)
#        else:
#            self.right_keypos = self.condition.left.position \
#                - len(left_sch)
#
#        # find left key position
#        if self.rightCondIsRightAttr:
#            self.left_keypos = self.condition.left.position
#        else:
#            self.left_keypos = self.condition.right.position
#
#        # define output tuple
#        outTuple = GrappaStagedTupleRef(gensym(), self.scheme())
#        out_tuple_type_def = outTuple.generateDefinition()
#        out_tuple_type = outTuple.getTupleTypename()
#        out_tuple_name = outTuple.name
#
#        # common index is defined by same right side and same key
#        # TODO: probably want also left side
#        hashtableInfo = state.lookupExpr((self.right, self.right_keypos))
#        if not hashtableInfo:
#            # if right child never bound then store hashtable symbol and
#            # call right child produce
#            self._hashname = self.__getHashName__()
#            _LOG.debug("generate hashname %s for %s", self._hashname, self)
#
#            hashname = self._hashname
#
#            # declaration of hash map
#            self.rightTupleTypeRef = state.createUnresolvedSymbol()
#            self.leftTupleTypeRef = state.createUnresolvedSymbol()
#            self.outTupleTypeRef = state.createUnresolvedSymbol()
#            right_type = self.rightTupleTypeRef.getPlaceholder()
#            left_type = self.leftTupleTypeRef.getPlaceholder()
#
#            # TODO: really want this addInitializers to be addPreCode
#            # TODO: *for all pipelines that use this hashname*
#            init_template = self._cgenv.get_template('hash_init.cpp')
#
#            state.addInitializers([init_template.render(locals())])
#            self.right.produce(state)
#
#            self.left.childtag = "left"
#            self.left.produce(state)
#
#            state.saveExpr((self.right, self.right_keypos),
#                           (self._hashname, right_type, left_type,
#                            self.right_syncname, self.left_syncname))
#
#        else:
#            # if found a common subexpression on right child then
#            # use the same hashtable
#            self._hashname, right_type, left_type,\
#                self.right_syncname, self.left_syncname = hashtableInfo
#            _LOG.debug("reuse hash %s for %s", self._hashname, self)
#
#        # now that Relation is produced,
#        # produce its contents by iterating over
#        # the join result
#        iterate_template = self._cgenv.get_template('result_scan.cpp')
#
#        hashname = self._hashname
#
#        state.addDeclarations([out_tuple_type_def])
#
#        pipeline_sync = create_pipeline_synchronization(state)
#        get_pipeline_task_name(state)
#
#        # add dependences on left and right inputs
#        state.addToPipelinePropertySet('dependences', self.right_syncname)
#        state.addToPipelinePropertySet('dependences', self.left_syncname)
#
#        # reduce is a single self contained pipeline.
#        # future hashjoin implementations may pipeline out of it
#        # by passing a continuation to reduceExecute
#        reduce_template = self._cgenv.get_template('reduce.cpp')
#
#        state.addPreCode(reduce_template.render(locals()))
#
#        delete_template = self._cgenv.get_template('delete.cpp')
#
#        state.addPostCode(delete_template.render(locals()))
#
#        inner_code_compiled = self.parent().consume(outTuple, self, state)
#
#        code = iterate_template % locals()
#        state.setPipelineProperty('type', 'in_memory')
#        state.setPipelineProperty('source', self.__class__)
#        state.addPipeline(code)

#    def consume(self, inputTuple, fromOp, state):
#        if fromOp.childtag == "right":
#            side = "Right"
#            self.right_syncname = get_pipeline_task_name(state)
#
#            keypos = self.right_keypos
#
#            self.rightTupleTypename = inputTuple.getTupleTypename()
#            if self.rightTupleTypeRef is not None:
#                state.resolveSymbol(self.rightTupleTypeRef,
#                                    self.rightTupleTypename)
#        elif fromOp.childtag == "left":
#            side = "Left"
#            self.left_syncname = get_pipeline_task_name(state)
#
#            keypos = self.left_keypos
#
#            self.leftTupleTypename = inputTuple.getTupleTypename()
#            if self.leftTupleTypeRef is not None:
#                state.resolveSymbol(self.leftTupleTypeRef,
#                                    self.leftTupleTypename)
#        else:
#            assert False, "src not equal to left or right"
#
#        hashname = self._hashname
#        keyname = inputTuple.name
#        keytype = inputTuple.getTupleTypename()
#        keyval = inputTuple.get_code(keypos)
#
#        # intra-pipeline sync
#        global_syncname = state.getPipelineProperty('global_syncname')
#
#        mat_template = self._cgenv.get_template('materialize.cpp')
#
#        # materialization point
#        code = mat_template.render(locals())
#        return code


class GrappaGroupBy(cppcommon.BaseCGroupby, GrappaOperator):
    _i = 0

    @classmethod
    def __genHashName__(cls):
        name = "%s_hash_%03d" % (cls.__name__, cls._i)
        cls._i += 1
        return name

    def __init__(self, *args):
        super(GrappaGroupBy, self).__init__(*args)
        self._cgenv = cppcommon.prepend_template_relpath(
            self.language().cgenv(),
            '{0}/groupby'.format(GrappaLanguage._template_path))

    def _combiner_for_builtin_update(self, update_op):
        # FIXME: should be using get_decomposable_state instead of this hack
        # FIXME: need AVG and STDEV
        if update_op.__class__ == aggregate.COUNT \
                or update_op.__class__ == aggregate.COUNTALL:
            return aggregate.SUM(update_op.input)
        else:
            return update_op

    def _init_func_for_op(self, op):
        r = {
            aggregate.MAX: 'std::numeric_limits<{state_type}>::lowest',
            aggregate.MIN: 'std::numeric_limits<{state_type}>::max'
        }.get(op.__class__)
        if r is None:
            return 'Aggregates::Zero<{state_type}>'
        else:
            return r

    def _reconstruct_aggregates_schema(self):
        """
         reconstruct the lost mapping of schema to aggregate
         expressions/grouping list
        """

        # TODO: doesn't this exist somewhere in raco?
        def resolve_name(ref, sch):
            if isinstance(
                    ref,
                    expression.UnnamedAttributeRef)or isinstance(
                    ref,
                    expression.UnnamedAttributeRef):
                return sch.get_names()[ref.position]
            else:
                return ref.name

        inp_sch = self.input.scheme()

        grouped_names = set([resolve_name(ref, inp_sch)
                             for ref in self.grouping_list])
        aggregates_types = [typ  # throw away the name because it is made up
                            for name, typ in self.scheme()
                            if name not in grouped_names]
        aggregates_names = [
            resolve_name(
                a.input,
                inp_sch) for a in self.aggregate_list]

        return scheme.Scheme(zip(aggregates_names, aggregates_types))

    def _assignment_code(self, output_tuple):
        """
         assign state type tuple to output type tuple
         For 0key case they should just be treated the same,
         i.e., self.state_tuple same as output_tuple for !self.useKey
         except for name
         """
        assignmentcode = ""
        for i in range(0, len(output_tuple.scheme)):
            d = output_tuple.set_func_code(i)
            s = output_tuple.get_code_with_name(
                i, "{0}_tmp".format(
                    output_tuple.name))
            assignment_template = self._cgenv.get_template('assignment.cpp')
            assignmentcode += assignment_template.render(
                dst_set_func=d, src_expr_compiled=s)

        return assignmentcode

    def _impl_produce(self, state):
        pipeline_sync = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        if self.useKey:
            inp_sch = self.input.scheme()

            numkeys = len(self.grouping_list)
            keytype = self._key_type(inp_sch)
            mapping_var_name = gensym()
            emit_type = self.state_tuple.getTupleTypename()

            # pass in attribute values individually
            produce_template = self._cgenv.get_template(
                'multi_uda_scan.cpp')
        else:
            produce_template = self._cgenv.get_template(
                'multi_uda_0key_output.cpp')

        output_tuple = GrappaStagedTupleRef(gensym(), self.scheme())
        output_tuple_name = output_tuple.name
        output_tuple_type = output_tuple.getTupleTypename()
        output_tuple_set_func = output_tuple.set_func_code(0)  # UNUSED??
        state.addDeclarations([output_tuple.generateDefinition()])

        inner_code = self.parent().consume(output_tuple, self, state)
        comment = self.language().comment("scan of " + str(self))

        assignmentcode = self._assignment_code(output_tuple)

        hashname = self._hashname
        state_type = self.state_tuple.getTupleTypename()
        combine_func = self.combine_func

        code = produce_template.render(locals())
        state.setPipelineProperty("type", "in_memory")
        state.addPipeline(code)

    def _key_type(self, inp_sch):
        return "std::tuple<{types}>".format(
            types=','.join([self.language().typename(
                g.typeof(inp_sch, None)) for g in self.grouping_list]))

    def _init_template(self):
        return self._cgenv.get_template('withkey_init.cpp')

    def _reuse_properties(self, other):
        self._hashname = other._hashname
        self.func_name = other.func_name
        self.state_tuple = other.state_tuple
        self.input_syncname = other.input_syncname
        self.input_type_ref = other.input_type_ref

    def produce(self, state):
        # we distinguish between no-key and using a key cases
        self.useKey = len(self.grouping_list) > 0
        _LOG.debug("groupby uses keys? %s" % self.useKey)

        # restrictions on aggregate support
        if not(self.useKey or
               all([not isinstance(exp, expression.UdaAggregateExpression)
                    for exp in self.aggregate_list])):
            raise NotImplementedError("""
                    UDAs with no groupby key. The reason is that we
                    need to support decomposable state for correctness
                    of a local aggregate and global combine strategy. This
                    is solved adhoc in specific important builtin cases
                    like COUNT""")

        if not all([not isinstance(exp, expression.ZeroaryOperator)
                    for exp in self.aggregate_list]):
            raise NotImplementedError("""No support for Zeroary aggregates yet.
            If using COUNT(*), then use COUNT(a), but NOTE that COUNT(a)
            does not have proper null semantics
            (unconditionally counts everything)""")

        self.aggregates_schema = self._reconstruct_aggregates_schema()

        hashtableInfo = state.lookupExpr(self)
        if hashtableInfo:
            subexpression_proxy = hashtableInfo
            self._reuse_properties(subexpression_proxy)
        else:
            symbol = gensym()
            self._hashname = GrappaGroupBy.__genHashName__()
            _LOG.debug("generate hashname %s for %s", self._hashname, self)
            self.func_name = "__{0}".format(symbol)

            self.state_tuple = GrappaStagedTupleRef(symbol,
                                                    self.aggregates_schema,
                                                    aligned=(not self.useKey))
            self.input_type_ref = state.createUnresolvedSymbol()
            state.saveExpr(self, self)
            state.addDeclarations([self.state_tuple.generateDefinition()])

        state_type = self.state_tuple.getTupleTypename()
        self.update_func = "{name}_update".format(name=self.func_name)
        self.init_func = "{name}_init".format(name=self.func_name)
        # combine_func currently just for 0key_output
        self.combine_func = "{name}_combine".format(name=self.func_name)
        init_func = self.init_func
        update_func = self.update_func
        combine_func = self.combine_func

        inp_sch = self.input.scheme()

        if self.useKey:
            numkeys = len(self.grouping_list)
            keytype = self._key_type(inp_sch)

        hashname = self._hashname

        if self.useKey:
            init_template = self._init_template()
            update_val_type = self.input_type_ref.getPlaceholder()
            valtype = state_type
            decl_template = self._cgenv.get_template('withkey_decl.cpp')
            state.addDeclarations([decl_template.render(locals())])
        else:
            no_key_state_initializer = \
                "symmetric_global_alloc<{state_tuple_type}>()".format(
                    state_tuple_type=self.state_tuple.getTupleTypename())

            init_template = self._cgenv.get_template('withoutkey_init.cpp')
            initializer = no_key_state_initializer
            func_name = self.func_name

        self.initializer = init_template.render(locals())
        state.addInitializers([self.initializer])

        if not hashtableInfo:
            self.input.produce(state)

        # now that everything is aggregated, produce the tuples

        # assert len(self.column_list()) == 1 \
        #    or isinstance(self.column_list()[0],
        #                  expression.AttributeRef), \
        #    """assumes first column is the key and second is aggregate result
#            column_list: %s""" % self.column_list()

        # add a dependence on the input aggregation pipeline
        for s in self.input_syncname:
            state.addToPipelinePropertySet('dependences', s)

        self._impl_produce(state)

    class AggregateSetter:

        def __init__(self, name, expression):
            self.name = name
            self.expression = expression

    def compile_assignments(self, assgns, inputTuple):
        """
        compile update statements
        """
        state_var_update_template = "auto {assignment};"
        state_var_updates = []
        state_vars = []
        decls = []
        inits = []

        for a in assgns:
            # doesn't have to use inputTuple.name,
            # but it will for simplicity
            # If it is a UDA, we need to compile the expression
            rhs = self.language().compile_expression(
                a.expression,
                # For resolving attributes of the consumed input tuple
                tupleref=inputTuple,
                # for resolving attributes of the state tuple
                state_scheme=self.aggregates_schema)

            # combine lhs, rhs with assignment
            code = "{lhs} = {rhs}".format(lhs=a.name, rhs=rhs[0])

            decls += rhs[1]
            inits += rhs[2]

            state_var_updates.append(
                state_var_update_template.format(assignment=code))
            state_vars.append(a.name)

        return state_var_updates, state_vars, decls, inits

    def _update_func(self, state, inputTuple):
        # add Builtins updaters and to those
        # from UDAs in self.updaters
        updaters_map = dict((k, (k, v)) for k, v in self.updaters)

        inp_sch = self.input.scheme()

        def aggregate_to_updater(index, aggr):
            if isinstance(aggr, aggregate.UdaAggregateExpression):
                attr_ref = aggr.input
                assert isinstance(
                    attr_ref, aggregate.NamedStateAttributeRef), \
                    "type(attr_ref)={0} but must be {1}".format(
                    type(attr_ref), aggregate.NamedStateAttributeRef.__class__)
                name, expr = updaters_map[attr_ref.name]
                return self.AggregateSetter(name, expr)
            elif isinstance(aggr, aggregate.BuiltinAggregateExpression):
                # get update function
                name = "_v{0}".format(index)
                input_type = self.language().typename(
                    aggr.input.typeof(inp_sch, None))
                op = aggr
                up_op_name = op.__class__.__name__
                state_type = self.language().typename(
                    aggr.typeof(
                        inp_sch,
                        None))
                update_func = \
                    "Aggregates::{op}<{state_type}, {input_type}>".format(
                        op=up_op_name,
                        state_type=state_type,
                        input_type=input_type)

                # create a class for this binary function so it can be compiled
                update_as_expression = expression.CustomBinaryFunction(
                    update_func,
                    # get the output type
                    self.aggregates_schema.get_types()[index],
                    # get the aggregate result attribute (left)
                    expression.NamedStateAttributeRef(
                        self.aggregates_schema.get_names()[index]),
                    # get the aggregate input attribute (right)
                    aggr.input)

                return self.AggregateSetter(name, update_as_expression)
            else:
                assert False, "expected every element of aggregate_list to " \
                              "be an aggregate"

        all_updaters = [
            aggregate_to_updater(
                i, a) for i, a in enumerate(
                self.aggregate_list)]

        update_updates, update_state_vars, update_decls, update_inits = \
            self.compile_assignments(all_updaters, inputTuple)

        update_def = self._cgenv.get_template(
            'update_definition.cpp').render(
            state_type=self.state_tuple.getTupleTypename(),
            input_type=inputTuple.getTupleTypename(),
            input_tuple_name=inputTuple.name,
            update_updates=update_updates,
            update_state_vars=update_state_vars,
            name=self.func_name)

        state.addDeclarations([update_def] + update_decls)
        state.addInitializers(update_inits)
        return update_state_vars

    def _init_def(self, state, inputTuple):
        # add Builtins inits and to those
        # from UDAs in self.inits
        inits_map = dict((k, (k, v)) for k, v in self.inits)
        inp_sch = self.input.scheme()

        def aggregate_to_init(index, aggr):
            if isinstance(aggr, aggregate.UdaAggregateExpression):
                attr_ref = aggr.input
                assert isinstance(
                    attr_ref, aggregate.NamedStateAttributeRef), \
                    "type(attr_ref)={0} but must be {1}".format(
                    type(attr_ref), aggregate.NamedStateAttributeRef.__class__)
                name, expr = inits_map[attr_ref.name]
                return self.AggregateSetter(name, expr)
            elif isinstance(aggr, aggregate.BuiltinAggregateExpression):
                name = "_v{0}".format(index)
                state_type = self.language().typename(
                    aggr.typeof(inp_sch, None))
                init_func = self._init_func_for_op(aggr) \
                    .format(state_type=state_type)
                return self.AggregateSetter(
                    name,
                    expression.CustomZeroaryFunction(
                        init_func,
                        # get the output type
                        self.aggregates_schema.get_types()[index]))

        all_initers = [
            aggregate_to_init(
                i, a) for i, a in enumerate(
                self.aggregate_list)]

        init_updates, init_state_vars, init_decls, init_inits = \
            self.compile_assignments(all_initers, inputTuple)

        init_def = self._cgenv.get_template('init_definition.cpp').render(
            state_type=self.state_tuple.getTupleTypename(),
            init_updates=init_updates,
            init_state_vars=init_state_vars,
            name=self.func_name)

        state.addDeclarations([init_def] + init_decls)
        state.addInitializers(init_inits)
        return init_state_vars

    def _combine_def(self, state, inputTuple):
        # add Builtins updaters and to those
        # from UDAs in self.updaters
        updaters_map = dict((k, (k, v)) for k, v in self.updaters)
        inp_sch = self.input.scheme()

        def aggregate_to_combiner(index, aggr):
            if isinstance(aggr, aggregate.UdaAggregateExpression):
                # TODO: Support decomposable aggregate state, instead of
                # TODO: just using the same function from self.updaters
                attr_ref = aggr.input
                assert isinstance(
                    attr_ref, aggregate.NamedStateAttributeRef), \
                    "type(attr_ref)={0} but must be {1}".format(
                    type(attr_ref), aggregate.NamedStateAttributeRef.__class__)
                name, expr = updaters_map[attr_ref.name]
                return self.AggregateSetter(name, expr)
            elif isinstance(aggr, aggregate.BuiltinAggregateExpression):
                # get combiner function
                name = "_v{0}".format(index)

                state_type = self.language().typename(
                    aggr.typeof(
                        inp_sch,
                        None))

                # hack to get the combiner function based on the aggregate
                # TODO: support decomposable state
                co_op_name = self._combiner_for_builtin_update(
                    aggr).__class__.__name__

                # type for left and right is state_type for combiner
                this_combine_func = "Aggregates::{op}<{type}, {type}>".format(
                    op=co_op_name, type=state_type)

                attribute0 = expression.NamedStateAttributeRef(
                    self.aggregates_schema.get_names()[index])
                attribute0.tagged_state_id = 0
                attribute1 = expression.NamedStateAttributeRef(
                    self.aggregates_schema.get_names()[index])
                attribute1.tagged_state_id = 1

                # create a class for this binary function so it can be compiled
                combine_as_expression = expression.CustomBinaryFunction(
                    this_combine_func,
                    # get the output type
                    self.aggregates_schema.get_types()[index],
                    # get the aggregate result attribute (left)
                    attribute0,
                    # get the aggregate input attribute (right)
                    attribute1)

                return self.AggregateSetter(name, combine_as_expression)
            else:
                assert False, "expected every element of aggregate_list to " \
                              "be an aggregate"

        all_combiners = [
            aggregate_to_combiner(
                i, a) for i, a in enumerate(
                self.aggregate_list)]

        combine_updates, combine_state_vars, combine_decls, combine_inits = \
            self.compile_assignments(all_combiners, inputTuple)

        # currently only needed for 0key reduce, since there is no key-based
        # reduce in the grappalang streaming aggregate
        combine_def = self._cgenv.get_template(
            'combine_definition.cpp').render(
            state_type=self.state_tuple.getTupleTypename(),
            combine_updates=combine_updates,
            combine_state_vars=combine_state_vars,
            name=self.func_name)

        state.addDeclarations([combine_def] + combine_decls)
        state.addInitializers(combine_inits)
        return combine_state_vars

    def _key_access_code(self, inputTuple, inp_sch):
        return [inputTuple.get_code(g.get_position(inp_sch))
                for g in self.grouping_list]

    def consume(self, inputTuple, fromOp, state):
        # save the inter-pipeline task name
        # Needs to be a list because could be multiple input occurences
        if not hasattr(self, 'input_syncname'):
            self.input_syncname = []
        self.input_syncname.append(get_pipeline_task_name(state))

        # not used, but must be resolved
        state.resolveSymbol(self.input_type_ref, inputTuple.getTupleTypename())

        inp_sch = self.input.scheme()

        update_vars = self._update_func(state, inputTuple)
        init_vars = self._init_def(state, inputTuple)

        assert set(update_vars) == set(init_vars), \
            """Initialized and update state vars are not the same \
            {0} != {1}""".format(update_vars, init_vars)

        if not self.useKey:
            # if 0key then add the combiner codes, otherwise
            # omit it because it may be invalid code
            combine_vars = self._combine_def(state, inputTuple)

            assert set(update_vars) == set(combine_vars), \
                """Combiner and update state vars are not the same \
                {0} != {1}""".format(update_vars,
                                     combine_vars)

        # generate the update and init function definitions
        state_tuple_decl = self.state_tuple.generateDefinition()

        # values for the materialize template (calling the update function)
        init_func = self.init_func
        update_func = self.update_func
        update_val = inputTuple.name
        input_type = inputTuple.getTupleTypename()

        if self.useKey:
            numkeys = len(self.grouping_list)
            keygets = self._key_access_code(inputTuple, inp_sch)

            materialize_template = self._cgenv.get_template('nkey_update.cpp')
        else:
            materialize_template = self._cgenv.get_template(
                'multi_uda_0key_update.cpp')

        if self.useKey:
            state.recordCodeWhenInLoop(self.language().comment("recycle") +
                                       "{hashname}->clear();\n".format(
                                           hashname=self._hashname))
        else:
            state.recordCodeWhenInLoop(self.language().comment("recycle") +
                                       self.initializer)

        hashname = self._hashname
        tuple_name = inputTuple.name
        pipeline_sync = state.getPipelineProperty("global_syncname")

        comment = self.language().comment("insert of " + str(self))

        code = materialize_template.render(locals())
        return code


def wait_statement(name):
    return GrappaLanguage.cgenv().get_template(
        'wait_statement.cpp').render(name=name)


def get_pipeline_task_name(state):
    name = "pipeline_{n}".format(n=state.getCurrentPipelineId())
    state.setPipelineProperty('sync', name)
    wait_stmt = wait_statement(name)
    state.addSeqWaitStatement(wait_stmt)
    return name


class GrappaHashJoin(GrappaJoin, GrappaOperator):
    _i = 0

    @staticmethod
    def __genHashName__():
        name = "hash_%03d" % GrappaHashJoin._i
        GrappaHashJoin._i += 1
        return name

    def __init__(self, *args):
        super(GrappaHashJoin, self).__init__(*args)
        self._cgenv = cppcommon.prepend_template_relpath(
            self.language().cgenv(),
            '{0}/hashjoin'.format(GrappaLanguage._template_path))

    def produce(self, state):
        declr_template = self._cgenv.get_template('hash_declaration.cpp')

        self.right.childtag = "right"

        my_sch = self.scheme()
        left_sch = self.left.scheme()
        right_sch = self.right.scheme()

        self.leftcols, self.rightcols = \
            algebra.convertcondition(self.condition,
                                     len(left_sch),
                                     left_sch + right_sch)

        keytype = CKeyUtils._aggregate_type(
            self.language(),
            right_sch,
            self.rightcols)

        # common index is defined by same right side and same key
        hashtableInfo = state.lookupExpr((self.right,
                                          frozenset(self.rightcols)))
        if not hashtableInfo:
            # if right child never bound then store hashtable symbol and
            # call right child produce
            self._hashname = GrappaHashJoin.__genHashName__()
            _LOG.debug("generate hashname %s for %s", self._hashname, self)

            hashname = self._hashname

            # declaration of hash map
            self.rightTupleTypeRef = state.createUnresolvedSymbol()
            in_tuple_type = self.rightTupleTypeRef.getPlaceholder()
            hashdeclr = declr_template.render(locals())
            state.addDeclarationsUnresolved([hashdeclr])

            init_template = self._cgenv.get_template('hash_init.cpp')

            self.initializer = init_template.render(locals())
            state.addInitializers([self.initializer])
            self.right.produce(state)
            state.saveExpr((self.right, frozenset(self.rightcols)),
                           (self._hashname, self.rightTupleTypename,
                            self.right_syncname))
            # TODO always safe here? I really want to call
            # TODO saveExpr before self.right.produce(),
            # TODO but I need to get the self.rightTupleTypename cleanly
        else:
            self.rightTupleTypeRef = None  # indicates CSE succeeded

            # if found a common subexpression on right child then
            # use the same hashtable
            self._hashname, self.rightTupleTypename, self.right_syncname\
                = hashtableInfo
            _LOG.debug("reuse hash %s for %s", self._hashname, self)

        self.left.childtag = "left"
        self.left.produce(state)

    def consume(self, t, src, state):
        if src.childtag == "right":
            comment = self.language().comment("right side of " + str(self))
            right_template = self._cgenv.get_template('insert_materialize.cpp')

            hashname = self._hashname
            keyname = t.name
            keyval = CKeyUtils._aggregate_val(t, self.rightcols)

            # Needs to be a list because could be multiple right sides
            # occurences
            if not hasattr(self, 'right_syncname'):
                self.right_syncname = []
            self.right_syncname.append(get_pipeline_task_name(state))

            self.rightTupleTypename = t.getTupleTypename()
            if self.rightTupleTypeRef is not None:
                state.resolveSymbol(self.rightTupleTypeRef,
                                    self.rightTupleTypename)

            pipeline_sync = state.getPipelineProperty('global_syncname')

            # materialization point
            code = right_template.render(locals())

            # recycling when right index is reused
            state.recordCodeWhenInLoop(
                self.language().comment("recycle") +
                "{hashname}.clear();\n".format(
                    hashname=hashname))

            return code

        if src.childtag == "left":
            comment = self.language().comment("left side of " + str(self))
            left_template = self._cgenv.get_template('lookup.cpp')

            # add a dependences on the right pipelines
            for s in self.right_syncname:
                state.addToPipelinePropertySet('dependences', s)

            hashname = self._hashname
            keyname = t.name
            input_tuple_type = t.getTupleTypename()
            keyval = CKeyUtils._aggregate_val(t, self.leftcols)

            pipeline_sync = state.getPipelineProperty('global_syncname')

            right_tuple_name = gensym()
            right_tuple_type = self.rightTupleTypename

            outTuple = GrappaStagedTupleRef(gensym(), self.scheme())
            out_tuple_type_def = outTuple.generateDefinition()
            out_tuple_type = outTuple.getTupleTypename()
            out_tuple_name = outTuple.name

            type1 = input_tuple_type
            type1numfields = len(t.scheme)
            type2 = right_tuple_type
            type2numfields = len(self.right.scheme())
            append_func_name, combine_function_def = \
                GrappaStagedTupleRef.get_append(
                    out_tuple_type,
                    type1, type1numfields,
                    type2, type2numfields)

            state.addDeclarations([out_tuple_type_def, combine_function_def])

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            code = left_template.render(locals())
            return code

        assert False, "src not equal to left or right"


def indentby(code, level):
    indent = " " * ((level + 1) * 6)
    return "\n".join([indent + line for line in code.split("\n")])

#
#
#
# class FreeMemory(GrappaOperator):
#  def fire(self, expr):
#    for ref in noReferences(expr)

# Basic selection like serial C++


class GrappaSelect(cppcommon.CBaseSelect, GrappaOperator):
    pass


# Basic apply like serial C++
class GrappaApply(cppcommon.CBaseApply, GrappaOperator):
    pass


# Basic duplication based bag union like serial C++
class GrappaUnionAll(cppcommon.CBaseUnionAll, GrappaOperator):
    pass


# Basic materialized copy based project like serial C++
class GrappaProject(cppcommon.CBaseProject, GrappaOperator):
    pass


class GrappaSink(cppcommon.CBaseSink, GrappaOperator):
    pass


class GrappaFileScan(cppcommon.CBaseFileScan, GrappaOperator):

    def new_tuple_ref_for_filescan(self, resultsym, scheme):
        # make new tuple 64-byte aligned if scanning into a global array
        aligned = self.array_representation == \
            _ARRAY_REPRESENTATION.GLOBAL_ARRAY
        return GrappaStagedTupleRef(resultsym, scheme,
                                    aligned=aligned)

    def __init__(self, representation=_ARRAY_REPRESENTATION.GLOBAL_ARRAY,
                 relation_key=None, _scheme=None, cardinality=None):
        self.array_representation = representation
        super(GrappaFileScan, self).__init__(
            relation_key, _scheme, cardinality)

    def __get_ascii_scan_template__(self):
        _LOG.warn("binary/ascii is command line choice")
        template_name = {
            _ARRAY_REPRESENTATION.GLOBAL_ARRAY: 'file_scan.cpp',
            _ARRAY_REPRESENTATION.SYMMETRIC_ARRAY:
                'symmetric_array_file_scan.cpp'
        }[self.array_representation]
        return self._language.cgenv().get_template(template_name)

    def __get_binary_scan_template__(self):
        _LOG.warn("binary/ascii is command line choice")
        template_name = {
            _ARRAY_REPRESENTATION.GLOBAL_ARRAY: 'file_scan.cpp',
            _ARRAY_REPRESENTATION.SYMMETRIC_ARRAY:
                'symmetric_array_file_scan.cpp'
        }[self.array_representation]
        return self._language.cgenv().get_template(template_name)

    def __get_relation_decl_template__(self, name):
        template_name = {
            _ARRAY_REPRESENTATION.GLOBAL_ARRAY:
                'global_array_relation_declaration.cpp',
            _ARRAY_REPRESENTATION.SYMMETRIC_ARRAY:
                'symmetric_array_relation_declaration.cpp'
        }[self.array_representation]

        return self._language.cgenv().get_template(template_name)

    def _get_input_aux_decls_template(self):
        return self._language.cgenv().get_template(
            'input_relation_declarations.cpp')

    def __repr__(self):
        return "{op}({rep!r}, {rk!r}, {sch!r}, {card!r})".format(
            rep=self.array_representation,
            op=self.opname(), rk=self.relation_key, sch=self._scheme,
            card=self._cardinality)


class GrappaStore(cppcommon.CBaseStore, GrappaOperator):

    def __file_code__(self, t, state):
        my_sch = self.scheme()

        filename = (str(self.relation_key).split(":")[2])
        outputnamedecl = define_cl_arg(
            'string',
            'output_file',
            '"{0}"'.format(filename),
            "Output File")

        state.addDeclarations([outputnamedecl])
        names = [x.encode('UTF8') for x in my_sch.get_names()]
        schemefile = \
            'writeSchema("{s}", FLAGS_output_file+".scheme");\n'.format(s=zip(
                names, my_sch.get_types()))
        state.addPreCode(schemefile)
        resultfile = 'writeTuplesUnordered(&result, FLAGS_output_file+".bin");'
        state.addPipelineFlushCode(resultfile)

        return ""


class TempRelationSymbol:

    def __init__(self, name):
        self._name = name

    def symbol(self):
        return "_temp_table_{}".format(self._name)

    def __eq__(self, other):
        return isinstance(other, TempRelationSymbol) \
            and self._name == other._name


class GrappaStoreTemp(algebra.StoreTemp, GrappaOperator):

    def produce(self, state):
        temp_rel = TempRelationSymbol(self.name)

        # Recycle cannot happen for every visit (in consume) because
        # it will get called with binary streaming operators.
        # Calling it here is safe because all pipelines below
        # the StoreTemp should not be using this temp relation
        #
        # An alternative approach to consider in the future is,
        # to just put recycle in consume() but require a barrier
        # after it that involves all pipelines this StoreTemp appears in

        self.num_producers = state.createUnresolvedSymbol()
        state.addCode(
            self._language.cgenv().get_template(
                'symmetric_array_temprelation_recycle.cpp').render(
                sym=temp_rel.symbol(),
                num_producers=self.num_producers.getPlaceholder()))

        self.input.produce(state)

    def consume(self, t, src, state):
        temp_rel = TempRelationSymbol(self.name)

        # uses symmetric array representation only

        tuple = state.lookupTupleDef(temp_rel.symbol())
        if not tuple:
            tuple = self.new_tuple_ref(temp_rel.symbol(), self.scheme())
            state.saveTupleDef(temp_rel.symbol(), tuple)

            declTupleType = tuple.generateDefinition()
            declRelation = self._language.cgenv()\
                .get_template('symmetric_array_temprelation_declaration.cpp')\
                .render(tuple_type=tuple.getTupleTypename(),
                        resultsym=temp_rel.symbol())

            state.addDeclarations([declTupleType,
                                   declRelation])

            init_stmt = self._language.cgenv()\
                .get_template('symmetric_array_temprelation_init.cpp')\
                .render(sym=temp_rel.symbol(),
                        tuple_type=tuple.getTupleTypename())

            state.addInitializers([init_stmt])

        newtuple = tuple.copy_type()
        state.addDeclarations([newtuple.generateDefinition()])
        assignment_code = \
            cppcommon.createTupleTypeConversion(self.language(),
                                                state,
                                                t,
                                                newtuple)

        state.addPostCode(self._language.cgenv()
                          .get_template(
            'symmetric_array_temprelation_materializer_done.cpp')
            .render(sym=temp_rel.symbol()))

        state.resolveCounterSymbol(self.num_producers)

        return assignment_code + self.language().comment(self) + self._language.cgenv()\
            .get_template('symmetric_array_temprelation_materialize.cpp')\
            .render(sym=temp_rel.symbol(), input_tuple_name=newtuple.name)


class GrappaNullInput(cppcommon.CBaseFileScan, GrappaOperator):

    def __get_binary_scan_template__(self):
        return None

    def __get_ascii_scan_template__(self):
        return None

    def produce(self, state):
        sym = TempRelationSymbol(self.relation_key)
        assert state.lookupTupleDef(sym.symbol()) is not None, \
            "Expected {} to already exist".format(sym.symbol())
        self.parent().consume(sym, self, state)

    def consume(self, t, src, state):
        assert False, "as a source, no need for consume"


class GrappaShuffle(algebra.Shuffle, GrappaOperator):

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        inner_plan_compiled = self.parent().consume(t, self, state)

        columnlist_nums = [c.position for c in self.columnlist]

        code = self.language().cgenv().get_template('shuffle.cpp').render(
            keyname=t.name,
            keytype=CKeyUtils._aggregate_type(
                self.language(),
                self.input.scheme(),
                columnlist_nums),
            keyval=CKeyUtils._aggregate_val(
                t,
                columnlist_nums),
            pipeline_sync=state.getPipelineProperty('global_syncname'),
            comment=self.language().comment(
                self.shortStr()),
            inner_code=inner_plan_compiled)

        return code


class MemoryScanOfFileScan(rules.Rule):

    def __init__(self, array_rep, memory_scan_class=GrappaMemoryScan):
        self._array_rep = array_rep
        self._memory_scan_class = memory_scan_class
        super(MemoryScanOfFileScan, self).__init__()

    """A rewrite rule for making a scan into materialization
     in memory then memory scan"""

    def fire(self, expr):
        if isinstance(expr, algebra.Scan) \
                and not isinstance(expr, GrappaFileScan):
            return self._memory_scan_class(
                GrappaFileScan(self._array_rep,
                               expr.relation_key,
                               expr.scheme()),
                self._array_rep)
        return expr

    def __str__(self):
        return "Scan => MemoryScan(FileScan) [{0}]".format(self._array_rep)


class ScanTempToMemoryScan(rules.Rule):

    """Logical ScanTemp to MemoryScan"""

    def fire(self, expr):
        if isinstance(expr, algebra.ScanTemp):
            return GrappaMemoryScan(GrappaNullInput(expr.name, expr.scheme()),
                                    _ARRAY_REPRESENTATION.SYMMETRIC_ARRAY,
                                    expr.name,
                                    expr.analyzed_num_tuples
                                    if hasattr(expr, 'analyzed_num_tuples')
                                    else 10000)

        return expr


class GrappaPartitionGroupBy(GrappaGroupBy):

    def __init__(self, *args):
        super(GrappaPartitionGroupBy, self).__init__(*args)
        # Override some GrappaGroupBy template with new ones.
        # The rest of the generation logic is the same
        self._cgenv = cppcommon.prepend_template_relpath(
            self._cgenv, '{0}/partition_groupby'.format(
                GrappaLanguage._template_path))


class GrappaBroadcastCrossProduct(algebra.CrossProduct, GrappaOperator):

    def produce(self, state):

        self.right.childtag = "right"
        self.right.produce(state)

        self.left.childtag = "left"
        self.left.produce(state)

    def _declare_broadcast_tuple(self, t, state):
        # declare global var and broadcast value
        self.broadcast_tuple = t.copy_type()
        var_decl = self.language().cgenv().get_template(
            'tuple_declaration.cpp').render(
            dst_type_name=self.broadcast_tuple.getTupleTypename(),
            dst_name=self.broadcast_tuple.name)
        state.addDeclarations([var_decl])

    def consume(self, t, src, state):
        if src.childtag == "right":
            # right to left dependency
            self.right_syncname = get_pipeline_task_name(state)

            code = self.language().comment(self.shortStr() + " RIGHT")

            self._declare_broadcast_tuple(t, state)

            code += """on_all_cores([=] {{
                  {global_name} = {input_name};
                   }});
                   """.format(global_name=self.broadcast_tuple.name,
                              input_name=t.name)

            return code

        elif src.childtag == "left":
            # right to left dependency
            state.addToPipelinePropertySet('dependences', self.right_syncname)

            code = self.language().comment(self.shortStr() + " LEFT")

            # add global field to your tuple

            output = GrappaStagedTupleRef(gensym(), self.scheme())

            type1 = t.getTupleTypename()
            type1numfields = len(t.scheme)
            type2 = self.broadcast_tuple.getTupleTypename()
            type2numfields = len(self.broadcast_tuple.scheme)
            append_func_name, combine_function_def = \
                GrappaStagedTupleRef.get_append(
                    output.getTupleTypename(),
                    type1, type1numfields,
                    type2, type2numfields)

            state.addDeclarations([output.generateDefinition(),
                                   combine_function_def])

            code += """
            {out_tuple_type} {out_tuple_name} =
              {append_func_name}({left_name}, {right_name});
              """.format(out_tuple_type=output.getTupleTypename(),
                         out_tuple_name=output.name,
                         append_func_name=append_func_name,
                         left_name=t.name,
                         right_name=self.broadcast_tuple.name)

            inner_plan_compiled = self.parent().consume(output, self, state)

            return code + inner_plan_compiled

        else:
            assert False, "bad childtag: {0}".format(src.childtag)


class Iterator(object):

    def operator_code(self, **kwargs):
        return self.iter_cgenv().get_template(
            "instantiate_operator.cpp").render(kwargs)

    def default_operator_code(self, **kwargs):
        kwargs['call_constructor'] = "{class_symbol}({inputsym})".format(
            **kwargs)
        return self.operator_code(**kwargs)

    def sink_operator_code(self, **kwargs):
        return self.iter_cgenv().get_template(
            "instantiate_sink.cpp").render(kwargs)

    def declare_sink(self, state, cgenv=None):
        if not cgenv:
            cgenv = self.iter_cgenv()
        symbol = "frag_{0}".format(gensym())
        state.setPipelineProperty('fragment_symbol', symbol)
        state.addDeclarations(
            [cgenv.get_template("sink_declaration.cpp").render(symbol=symbol)])
        return symbol

    def assign_symbol(self):
        self.symbol = gensym()

    def iter_cgenv(cls, env=GrappaLanguage.cgenv()):
        return cppcommon.prepend_template_relpath(
            env, '{0}/iterators/'.format(GrappaLanguage._template_path))


class IGrappaMemoryScan(GrappaMemoryScan, Iterator):

    def _constructor(self, inputsym, state):
        return "Scan<{type}>({inputsym})".format(
            type=state.lookupTupleDef(inputsym).getTupleTypename(),
            inputsym=inputsym)

    def consume(self, inputsym, src, state):
        _ = create_pipeline_synchronization(state)
        _ = get_pipeline_task_name(state)

        self.assign_symbol()
        stagedTuple = state.lookupTupleDef(inputsym)
        state.addOperator(self.operator_code(
            produce_type=stagedTuple.getTupleTypename(),
            symbol=self.symbol,
            call_constructor=self._constructor(inputsym, state)
        ))
        self.parent().consume(stagedTuple, self, state)

        state.addPipeline()
        return None


class IGrappaSelect(GrappaSelect, Iterator):

    def consume(self, t, src, state):
        class_symbol = "Select_{sym}".format(sym=gensym())
        self.assign_symbol()
        state.addDeclarations([self.iter_cgenv().get_template(
            "select.cpp").render(
            class_symbol=class_symbol,
            consume_type=t.getTupleTypename(),
            produce_type=t.getTupleTypename(),
            consume_tuple_name=t.name,
            expression=self._compile_condition(t, state)
        )])
        state.addOperator(self.default_operator_code(
            produce_type=t.getTupleTypename(),
            symbol=self.symbol,
            inputsym=src.symbol,
            class_symbol=class_symbol
        ))
        self.parent().consume(t, self, state)
        return None


class IGrappaStore(GrappaStore, Iterator):

    def _constructor(self, t, inputsym, state):
        return "Store<{type}>({inputsym}, &result)".format(
            type=t.getTupleTypename(),
            inputsym=inputsym)

    def consume(self, t, src, state):
        symbol = self.declare_sink(state)
        self._add_result_declaration(t, state)
        state.addOperator(self.sink_operator_code(
            symbol=symbol,
            call_constructor=self._constructor(t, src.symbol, state)
        ))
        return None


class IGrappaApply(GrappaApply, Iterator):

    def consume(self, t, src, state):
        class_symbol = "Apply_{}".format(gensym())
        self.symbol = gensym()

        state.addDeclarations([self.iter_cgenv().get_template(
            "apply.cpp").render(
            class_symbol=class_symbol,
            produce_type=self.newtuple.getTupleTypename(),
            consume_type=t.getTupleTypename(),
            produce_tuple_name=self.newtuple.name,
            consume_tuple_name=t.name,
            statements=self._apply_statements(t, state)
        )])
        state.addOperator(self.default_operator_code(
            produce_type=self.newtuple.getTupleTypename(),
            class_symbol=class_symbol,
            symbol=self.symbol,
            inputsym=src.symbol
        ))

        self.parent().consume(self.newtuple, self, state)

        return None


class IGrappaHashJoin(GrappaSymmetricHashJoin, Iterator):

    def _constructor(self, inputsym, class_symbol):
        return "{class_symbol}(&{hashname}, {inputsym})".format(
            class_symbol=class_symbol,
            hashname=self._hashname,
            inputsym=inputsym
        )

    def produce(self, state):
        super(IGrappaHashJoin, self).produce(state)

        _ = create_pipeline_synchronization(state)
        _ = get_pipeline_task_name(state)

        # add dependences on left and right inputs
        state.addToPipelinePropertySet('dependences', self.right_syncname)
        state.addToPipelinePropertySet('dependences', self.left_syncname)

        self.assign_symbol()
        class_symbol = "HashJoinSource_{sym}".format(sym=gensym())
        keytype = CKeyUtils._aggregate_type(
            self.language(),
            self.right.scheme(),
            self.rightcols)
        left_type = self.leftTypeRef.getPlaceholder()
        right_type = self.rightTypeRef.getPlaceholder()
        out_tuple_type = self.outTuple.getTupleTypename()

        append_func_name, combine_function_def = \
            GrappaStagedTupleRef.get_append(
                out_tuple_type,
                left_type, len(self.left.scheme()),
                right_type, len(self.right.scheme()))

        state.addDeclarations([combine_function_def])

        state.addDeclarations([self.iter_cgenv().get_template(
            "hashjoin_source.cpp").render(
            class_symbol=class_symbol,
            keytype=keytype,
            left_tuple_type=left_type,
            right_tuple_type=right_type,
            out_tuple_type=out_tuple_type,
            left_name=gensym(),
            right_name=gensym(),
            append_func_name=append_func_name
        )])

        state.addOperator(self.operator_code(
            produce_type=out_tuple_type,
            symbol=self.symbol,
            call_constructor="{class_symbol}(&{hashname})".format(
                class_symbol=class_symbol,
                hashname=self._hashname)
        ))

        self.parent().consume(self.outTuple, self, state)
        state.addPipeline()

    def consume(self, t, src, state):
        class_symbol_template = "HashJoinSink{side}_{sym}"
        symbol = self.declare_sink(state)

        if src.childtag == 'right':
            side = 'Right'
            class_symbol = class_symbol_template.format(
                side=side,
                sym=gensym())

            keyval = CKeyUtils._aggregate_val(t, self.rightcols)
            keytype = CKeyUtils._aggregate_type(
                self.language(),
                self.right.scheme(),
                self.rightcols)
            state.resolveSymbol(self.rightTypeRef, t.getTupleTypename())

            # save to add after left type added
            self.right_class_decl = self.iter_cgenv().get_template(
                "hashjoin_sink.cpp").render(
                class_symbol=class_symbol,
                side=side,
                keytype=keytype,
                left_tuple_type=self.leftTypeRef.getPlaceholder(),
                right_tuple_type=t.getTupleTypename(),
                input_tuple_name=t.name,
                keyval=keyval,
                pipeline_sync=state.getPipelineProperty("global_syncname"))

            self.right_syncname = get_pipeline_task_name(state)

        elif src.childtag == 'left':
            side = 'Left'
            class_symbol = class_symbol_template.format(
                side=side,
                sym=gensym())

            keyval = CKeyUtils._aggregate_val(t, self.leftcols)
            keytype = CKeyUtils._aggregate_type(
                self.language(),
                self.left.scheme(),
                self.leftcols)
            state.resolveSymbol(self.leftTypeRef, t.getTupleTypename())

            state.addDeclarations([self.iter_cgenv().get_template(
                "hashjoin_sink.cpp").render(
                class_symbol=class_symbol,
                side=side,
                keytype=keytype,
                left_tuple_type=t.getTupleTypename(),
                right_tuple_type=self.rightTypeRef.getPlaceholder(),
                input_tuple_name=t.name,
                keyval=keyval,
                pipeline_sync=state.getPipelineProperty("global_syncname")
                # add right side here now that left type declared
            ), self.right_class_decl])

            self.left_syncname = get_pipeline_task_name(state)

        else:
            assert False, "Invalid child tag: {}".format(src.childtag)

        state.addOperator(self.sink_operator_code(
            symbol=symbol,
            call_constructor=self._constructor(src.symbol, class_symbol)
        ))

        return None


class IGrappaGroupBy(GrappaGroupBy, Iterator):

    def __init__(self, *args):
        super(IGrappaGroupBy, self).__init__(*args)
        self._igroupby_cgenv = self.iter_cgenv(env=self._cgenv)

    def _reuse_properties(self, other):
        super(IGrappaGroupBy, self)._reuse_properties(other)
        # save additional properties
        self.input_tuple_type = other.input_tuple_type

    def _impl_produce(self, state):
        pipeline_sync = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        if self.useKey:
            produce_template = self._igroupby_cgenv.get_template(
                "multikey_groupby_source.cpp")
            class_symbol = "AggregateSource_{}".format(gensym())
        else:
            produce_template = self._igroupby_cgenv.get_template(
                "0key_groupby_source.cpp")
            class_symbol = "ZeroKeyAggregateSource_{}".format(gensym())

        output_tuple = GrappaStagedTupleRef(gensym(), self.scheme())
        state.addDeclarations([output_tuple.generateDefinition()])

        assignment_code = self._assignment_code(output_tuple)

        self.assign_symbol()

        inp_sch = self.input.scheme()

        state.addDeclarations([produce_template.render(
            class_symbol=class_symbol,
            produce_type=output_tuple.getTupleTypename(),
            produce_tuple_name=output_tuple.name,
            keytype=self._key_type(inp_sch),
            state_type=self.state_tuple.getTupleTypename(),
            input_type=self.input_tuple_type,
            assignment_code=assignment_code,
            mapping_var_name=output_tuple.name + "_entry",
            combine_func=self.combine_func,
            state_tuple_name=output_tuple.name + "_tmp"
        )])

        state.addOperator(
            self.operator_code(
                produce_type=output_tuple.getTupleTypename(),
                symbol=self.symbol,
                call_constructor="{class_symbol}({hashname})".format(
                    class_symbol=class_symbol,
                    hashname=self._hashname)))

        self.parent().consume(output_tuple, self, state)

        state.setPipelineProperty("type", "in_memory")
        state.addPipeline()

    def _init_template(self):
        return self._igroupby_cgenv.get_template('withkey_init.cpp')

    def consume(self, t, src, state):
        # save the inter-pipeline task name
        if not hasattr(self, 'input_syncname'):
            self.input_syncname = []
        self.input_syncname.append(get_pipeline_task_name(state))

        self.input_tuple_type = t.getTupleTypename()

        state.resolveSymbol(self.input_type_ref, self.input_tuple_type)

        inp_sch = self.input.scheme()

        self._update_func(state, t)
        self._init_def(state, t)

        # generate class definition, if needed
        if self.useKey:
            class_symbol = "AggregateSink_{}".format(gensym())
            state.addDeclarations([
                self._igroupby_cgenv.get_template(
                    "multikey_groupby_sink.cpp").render(
                    class_symbol=class_symbol,
                    consume_type=t.getTupleTypename(),
                    keytype=self._key_type(inp_sch),
                    state_type=self.state_tuple.getTupleTypename(),
                    pipeline_sync=state.getPipelineProperty("global_syncname"),
                    consume_tuple_name=t.name,
                    keygets=self._key_access_code(t, inp_sch)
                )
            ])
        else:
            self._combine_def(state, t)

        symbol = self.declare_sink(state, cgenv=self._igroupby_cgenv)

        # generate instantiation
        if self.useKey:
            state.addOperator(
                self.sink_operator_code(
                    symbol=symbol,
                    call_constructor="""
                    {class_symbol}({inputsym}, {hashname})""".format(
                        inputsym=src.symbol,
                        class_symbol=class_symbol,
                        hashname=self._hashname)))
        else:
            state.addOperator(
                self.sink_operator_code(
                    symbol=symbol,
                    call_constructor="""
                    ZeroKeyAggregateSink<{consume_type}, {state_type}>(
                    {inputsym}, {hashname}, &{update_func}, &{init_func})
                    """.format(
                        inputsym=src.symbol,
                        hashname=self._hashname,
                        update_func=self.update_func,
                        init_func=self.init_func,
                        consume_type=t.getTupleTypename(),
                        state_type=self.state_tuple.getTupleTypename())))

        return None


class IGrappaPartitionGroupBy(IGrappaGroupBy):

    def __init__(self, *args):
        super(IGrappaPartitionGroupBy, self).__init__(*args)
        # Override some IGrappaGroupBy template with new ones.
        # The rest of the generation logic is the same
        self._igroupby_cgenv = cppcommon.prepend_template_relpath(
            self._igroupby_cgenv,
            '{0}/iterators/partition_groupby'.format(
                GrappaLanguage._template_path))


class IGrappaShuffle(algebra.Shuffle, GrappaOperator, Iterator):

    def __init__(self, *args):
        raise NotImplementedError(
            "Should not be necessary currently, see {0}".format(
                FuseGroupByShuffle))

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        # currently a no-op because IGroupBy takes care of the shuffle
        self.parent().consume(t, src, state)
        return None


class IGrappaBroadcastCrossProduct(GrappaBroadcastCrossProduct, Iterator):

    def consume(self, t, src, state):
        if src.childtag == "right":
            symbol = self.declare_sink(state)
            # right to left dependency
            self.right_syncname = get_pipeline_task_name(state)
            self._declare_broadcast_tuple(t, state)
            state.addOperator(
                self.sink_operator_code(
                    symbol=symbol,
                    call_constructor="""BroadcastTupleSink<{consume_type}>
                    ({inputsym}, &{broadcast_tuple})""".format(
                        consume_type=t.getTupleTypename(),
                        inputsym=src.symbol,
                        broadcast_tuple=self.broadcast_tuple.name)))
            return None

        elif src.childtag == "left":
            # right to left dependency
            state.addToPipelinePropertySet('dependences', self.right_syncname)
            output = GrappaStagedTupleRef(gensym(), self.scheme())

            append_func_name, combine_function_def = \
                GrappaStagedTupleRef.get_append(
                    output.getTupleTypename(),
                    t.getTupleTypename(),
                    len(t.scheme),
                    self.broadcast_tuple.getTupleTypename(),
                    len(self.broadcast_tuple.scheme))

            class_symbol = "BroadcastTupleStream_{}".format(gensym())

            class_decl = self.iter_cgenv().get_template(
                'broadcast_stream.cpp').render(
                class_symbol=class_symbol,
                left_type=t.getTupleTypename(),
                right_type=self.broadcast_tuple.getTupleTypename(),
                output_type=output.getTupleTypename(),
                output_name=output.name,
                append_func_name=append_func_name)

            state.addDeclarations([output.generateDefinition(),
                                   combine_function_def,
                                   class_decl])

            self.assign_symbol()
            state.addOperator(
                self.operator_code(
                    produce_type=output.getTupleTypename(),
                    symbol=self.symbol,
                    call_constructor="""
                    {class_symbol}({inputsym}, &{broadcast_tuple})""".format(
                        class_symbol=class_symbol,
                        broadcast_tuple=self.broadcast_tuple.name,
                        inputsym=src.symbol)))

            self.parent().consume(output, self, state)
            return None

        else:
            assert False, "bad childtag: {0}".format(src.childtag)


class GrappaSequence(algebra.Sequence, GrappaOperator):

    def produce(self, state):
        for c in self.children():
            c.produce(state)

            # During c.produce, the compiler will be adding
            # wait statements for each pipeline in the child c.
            # To enforce sequential execution of the
            # sequence, we need to end each child with
            # waits for its pipelines
            waits = state.getAndFlushSeqWaitStatements()
            for s in waits:
                state.addCode(s)

    def consume(self, sym, src, state):
        raise NotImplementedError(
            "GrappaSequence is a statement not expression. Its children"
            "should not call consume")


class GrappaDoWhile(algebra.DoWhile, GrappaOperator):

    def produce(self, state):
        state.addDeclarations(["bool continue_while;"])

        state.addCode("""
        WhileLoopManager while_manager;
        do {
        while_manager.iteration_start();
        continue_while = false;
        auto continue_while_a = make_global(&continue_while);
        """)

        state.enterLoop()

        for c in self.args:  # condition (args[-1]) is not treated specially
            c.produce(state)

            # See GrappaSequence.produce() for explanation
            waits = state.getAndFlushSeqWaitStatements()
            for s in waits:
                state.addCode(s)

        # put the recycle code before the pipeline code
        # within the loop, pipelines are queued instead of emitted
        pipes, recycles = state.exitLoop()
        for c in recycles:
            state.addCode(c)
        for c in pipes:
            state.addCode(c)

        state.addCode("""
            while_manager.iteration_end();
        } while (continue_while);
        """)

    def consume(self, sym, src, state):
        raise NotImplementedError(
            "GrappaDoWhile is a statement not expression. Its children"
            "should not call consume")


class GrappaTest(algebra.UnaryOperator, GrappaOperator):

    """Sets the found "register" to true if relation is
    non-empty"""

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        return """delegate::write(continue_while_a, {get});
        """.format(get=t.get_code(0))

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def shortStr(self):
        return "{op}".format(op=self.opname())


class GrappaWhileCondition(rules.Rule):

    """In the DoWhile condition, insert a test operator to check
    result of the condition table"""

    def fire(self, expr):
        if isinstance(expr, algebra.DoWhile):
            newdw = rules.OneToOne(algebra.DoWhile, GrappaDoWhile).fire(expr)
            newdw.copy(expr)
            newdw.children()[-1] = GrappaTest(expr.children()[-1])
            return newdw
        return expr

    def __str__(self):
        return """DoWhile[..., Cond] => \
        GrappaDoWhile[..., GrappaTest[Cond]]"""


class DistinctToGroupby(rules.Rule):
    """Change distinct back to Groupby(attributes)"""

    def __init__(self, groupby_class):
        self.gb_class = groupby_class

    def fire(self, expr):
        if isinstance(expr, algebra.Distinct):
            return self.gb_class([expression.UnnamedAttributeRef(i)
                                  for i in range(len(expr.scheme()))],
                                 [],
                                 expr.input)
        else:
            return expr

    def __str__(self):
        return """Distinct => Groupby(all attributes, no aggregate)"""


class CrossProductWithSmall(rules.Rule):

    """
    If there is a cross product between a relation and
    a singleton, then broadcast the singleton: cross product
    is just adding an attribute to every tuple from the other relation
    """

    def __init__(
            self,
            broadcast_crossproduct_class=GrappaBroadcastCrossProduct):
        self.op_class = broadcast_crossproduct_class

    def fire(self, expr):
        if isinstance(expr, algebra.CrossProduct):
            if expr.right.num_tuples() == 1:
                return self.op_class(expr.left, expr.right)

        return expr

    def __str__(self):
        return "CrossProduct(big, singleton) => " \
               "{}(big, singleton)".format(self.op_class.__name__)


class FuseGroupByShuffle(rules.Rule):

    """
    Global groupby is a fusion of a shuffle and a groupby
    """

    def __init__(self, global_groupby_class):
        self.global_groupby_class = global_groupby_class

    def fire(self, expr):
        if isinstance(
                expr,
                algebra.GroupBy) and isinstance(
                expr.input,
                algebra.Shuffle):
            # make sure the shuffle is for the groupby
            if expr.grouping_list == expr.input.columnlist:
                x = rules.OneToOne(
                    algebra.GroupBy,
                    self.global_groupby_class).fire(expr)
                x.input = expr.input.input
                return x
            else:
                _LOG.warning(
                    "missed {0} because the shuffle isn't for the groupby")
        else:
            return expr

    def __str__(self):
        return "GroupBy($a..$z)[Shuffle($a..$z)] => {0}($a..$z)".format(
            self.global_groupby_class)


def iteratorfy(emit_print, scan_array_repr, groupby_class):
    return [
        FuseGroupByShuffle(IGrappaGroupBy),

        rules.ProjectingJoinToProjectOfJoin(),

        rules.OneToOne(algebra.Select, IGrappaSelect),
        MemoryScanOfFileScan(scan_array_repr, IGrappaMemoryScan),
        rules.OneToOne(algebra.Apply, IGrappaApply),
        rules.OneToOne(algebra.Join, IGrappaHashJoin),
        # rules.OneToOne(algebra.Shuffle, IGrappaShuffle),
        # rules.OneToOne(algebra.Project, GrappaProject),
        # rules.OneToOne(algebra.UnionAll, GrappaUnionAll),
        cppcommon.StoreToBaseCStore(emit_print, IGrappaStore),
        CrossProductWithSmall(IGrappaBroadcastCrossProduct),
        DistinctToGroupby(groupby_class)
    ]


def grappify(join_type, emit_print,
             scan_array_repr, groupby_class):
    if isinstance(join_type, str):
        join_type_class = dict(
            (c.__name__,
             c) for c in [
                GrappaHashJoin,
                GrappaSymmetricHashJoin,
                GrappaShuffleHashJoin,
                GrappaOverSynchronizedSymmetricHashJoin])[join_type]
    else:
        join_type_class = join_type

    return [
        rules.ProjectingJoinToProjectOfJoin(),

        rules.OneToOne(algebra.Select, GrappaSelect),
        MemoryScanOfFileScan(scan_array_repr),
        rules.OneToOne(algebra.Apply, GrappaApply),
        rules.OneToOne(algebra.Join, join_type_class),
        rules.OneToOne(algebra.Project, GrappaProject),
        rules.OneToOne(algebra.Shuffle, GrappaShuffle),
        rules.OneToOne(algebra.UnionAll, GrappaUnionAll),
        cppcommon.StoreToBaseCStore(emit_print, GrappaStore),
        CrossProductWithSmall(),
        rules.OneToOne(algebra.Sink, GrappaSink),
        rules.OneToOne(algebra.StoreTemp, GrappaStoreTemp),
        ScanTempToMemoryScan(),
        rules.OneToOne(algebra.Sequence, GrappaSequence),
        rules.OneToOne(algebra.DoWhile, GrappaDoWhile),
        rules.OneToOne(algebra.SingletonRelation, GrappaSingletonRelation),
        DistinctToGroupby(groupby_class)

        # Don't need this because we support two-key
        # cppcommon.BreakHashJoinConjunction(GrappaSelect, join_type)
    ]


class GrappaAlgebra(Algebra):

    def __init__(self, emit_print=cppcommon.EMIT_CONSOLE):
        self.emit_print = emit_print

    def opt_rules(self,
                  join_type=GrappaHashJoin,
                  scan_array_repr=_ARRAY_REPRESENTATION.GLOBAL_ARRAY,
                  compiler='push',
                  SwapJoinSides=False,
                  external_indexing=False,
                  **kwargs):

        # datalog_rules = [
        # rules.removeProject(),
        #     rules.CrossProduct2Join(),
        #     rules.SimpleGroupBy(),
        # SwapJoinSides(),
        #     rules.OneToOne(algebra.Select, GrappaSelect),
        #     rules.OneToOne(algebra.Apply, GrappaApply),
        # rules.OneToOne(algebra.Scan,MemoryScan),
        #     MemoryScanOfFileScan(),
        # rules.OneToOne(algebra.Join, GrappaSymmetricHashJoin),
        #     rules.OneToOne(algebra.Join, self.join_type),
        #     rules.OneToOne(algebra.Project, GrappaProject),
        #     rules.OneToOne(algebra.GroupBy, GrappaGroupBy),
        # TODO: this Union obviously breaks semantics
        #     rules.OneToOne(algebra.Union, GrappaUnionAll),
        #     rules.OneToOne(algebra.Store, GrappaStore)
        # rules.FreeMemory()
        # ]

        if compiler == 'push':
            groupby_classes = {
                'global': GrappaGroupBy,
                'partition': GrappaPartitionGroupBy}
        elif compiler == 'iterator':
            groupby_classes = {
                'global': IGrappaGroupBy,
                'partition': IGrappaPartitionGroupBy}
        else:
            raise ValueError(
                "unsupported argument compiler={0}".format(compiler))

        groupby_sematics = kwargs.get('groupby_semantics', 'global')
        if groupby_sematics == 'global':
            groupby_rules = [
                rules.OneToOne(
                    algebra.GroupBy,
                    groupby_classes['global'])]
            groupby_class = groupby_classes['global']
        elif groupby_sematics == 'partition':
            groupby_rules = rules.distributed_group_by(
                groupby_classes['partition'],
                countall_rule=False,
                only_fire_on_multi_key=groupby_classes['global'])
            groupby_class = groupby_classes['partition']
        else:
            raise ValueError(
                "groupby_semantics must be one of {global, partition}")

        if compiler == 'push':
            grappify_rules = grappify(
                join_type,
                self.emit_print,
                scan_array_repr, groupby_class)
        elif compiler == 'iterator':
            grappify_rules = iteratorfy(self.emit_print, scan_array_repr,
                                        groupby_class)

        # sequence that works for myrial
        rule_grps_sequence = [
            rules.remove_trivial_sequences,
            [GrappaWhileCondition()],
            rules.simple_group_by,
            cppcommon.clang_push_select,
            rules.push_project,
            rules.push_apply,
            groupby_rules,
            [rules.NumTuplesPropagation()],
            [rules.DeDupBroadcastInputs()],
            grappify_rules
        ]

        # switch join order
        if SwapJoinSides:
            rule_grps_sequence.insert(0, [rules.SwapJoinSides()])

        # set external indexing on (replacing strings with ints)
        if external_indexing:
            CBaseLanguage.set_external_indexing(True)

        # flatten the rules lists
        rule_list = list(itertools.chain(*rule_grps_sequence))

        # disable specified rules
        rules.Rule.apply_disable_flags(rule_list, *kwargs.keys())

        return rule_list




import collections

import raco.algebra
import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco import types


class FileScanTest(myrial_test.MyrialTestCase):

    def test_filescan(self):
        query = """
x = load("examples/load_options.csv",
csv(
    schema(column0:int, column1:string, column2:string, column3:float),
    delimiter="|", quote="~", escape="%", skip=2));
store(x, OUTPUT);
        """
        expected = collections.Counter([
            (1, "foo", "abc|def", 1.0),
            (2, "bar", "ghi|jkl", 2.0),
        ])
        self.check_result(query, expected)



# coding=utf-8
""" Tests for expressions with UDFs """
from collections import Counter

from python_test import PythonTestCase
from raco.algebra import Apply
from raco.backends.myria.connection import FunctionTypes
from raco.python import convert
from raco.python.exceptions import PythonArgumentException
from raco.types import STRING_TYPE, BOOLEAN_TYPE, LONG_TYPE, INT_TYPE


class TestUDF(PythonTestCase):
    def _execute(self, udfs, query, expected):
        projection = convert(query, [self.schema], udfs)
        self.assertIsNotNone(projection)

        expression = Apply([('out', projection)], self.scan)
        plan = self.get_query(expression)
        return self.check_result(plan, expected)

    @staticmethod
    def _make_udf(name, source, out_type, arity):
        return {'name': name,
                'source': source,
                'outputType': out_type,
                'inputSchema': [INT_TYPE for _ in xrange(arity)],
                'lang': FunctionTypes.PYTHON}

    def test_invocation(self):
        self._execute(
            [self._make_udf('udf', 'lambda i: i == 5', BOOLEAN_TYPE, 1)],
            """lambda t: udf(t[0])""",
            Counter([(0,)] * 6 + [(1,)]))

    def test_invocation_in_expression(self):
        self._execute(
            [self._make_udf('udf', 'lambda i: i == 5', LONG_TYPE, 1)],
            """lambda t: udf(t[0]) + 1""",
            Counter([(1,)] * 6 + [(2,)]))

    def test_multiple_invocation(self):
        self._execute(
            [self._make_udf('udf', 'lambda i: i == 5', LONG_TYPE, 1),
             self._make_udf('udf2', 'lambda i: i < 10', LONG_TYPE, 1)],
            """lambda t: udf(t[0]) + udf2(t[0])""",
            Counter([(1,)] * 6 + [(2,)]))

    def test_unknown_udf(self):
        self.assertRaises(PythonArgumentException,
                          lambda: self._execute(
                              [self._make_udf('udf', 'lambda i: i == 5',
                                              LONG_TYPE, 1)],
                              """lambda t: udffoo(t[0])""",
                              Counter([(1,)] * 6 + [(2,)])))

    def test_arity(self):
        self._execute(
            [self._make_udf('udf', 'lambda i,j: i == j', LONG_TYPE, 2)],
            """lambda t: udf(t[0], t[0])""",
            Counter([(1,)] * len(self.emp_table)))

    def test_string_parameter(self):
        self._execute(
            [self._make_udf('udf', 'lambda s: s[0] == "B"', LONG_TYPE, 1)],
            """lambda t: udf(t[2])""",
            Counter([(0,)] * 6 + [(1,)]))

    def test_string_return(self):
        self._execute(
            [self._make_udf('udf', 'lambda s: s[0]', STRING_TYPE, 1)],
            """lambda t: udf(t[2])""",
            Counter({('D',): 2, ('A',): 1, ('V',): 1, ('B',): 1,
                    ('M',): 1, ('S',): 1}))



from abc import abstractmethod, ABCMeta
from ast import literal_eval
import os
import json

from raco.algebra import DEFAULT_CARDINALITY
from raco.representation import RepresentationProperties
from raco.relation_key import RelationKey
from raco.scheme import Scheme


class Relation(object):

    def __init__(self, name, sch):
        self.name = name
        self._scheme = sch

    def __eq__(self, other):
        return self.name == other.name
        # and self.scheme == other.scheme

    def scheme(self):
        return self._scheme


class FileRelation(Relation):
    pass


class ASCIIFile(FileRelation):
    pass


class Catalog(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_num_servers(self):
        """ Return number of servers in myria deployment """

    @abstractmethod
    def get_scheme(self, rel_key):
        """ Return scheme of tuples of rel_key """

    @abstractmethod
    def num_tuples(self, rel_key):
        """ Return number of tuples of rel_key """

    @abstractmethod
    def partitioning(self, rel_key):
        """ Return partitioning of rel_key """

    def representation_properties(self, rel_key):
        """
        Return interesting properties, like partitioning and sorted
        """
        # default is to return no information
        return RepresentationProperties()


# Some useful Catalog implementations
class FakeCatalog(Catalog):
    """ fake catalog, should only be used in test """

    def __init__(self, num_servers, child_sizes=None,
                 child_partitionings=None, child_functions=None):
        self.num_servers = num_servers
        # default sizes
        self.sizes = {}
        # default partitionings
        self.partitionings = {}
        # overwrite default sizes if necessary
        self.functions = {}

        if child_sizes:
            for child, size in child_sizes.items():
                self.sizes[RelationKey(child)] = size
        if child_partitionings:
            for child, part in child_partitionings.items():
                self.partitionings[RelationKey(child)] = tuple(part)
        if child_functions:
            for child, typ in child_functions.items():
                self.functions[child] = funcObj

    def get_num_servers(self):
        return self.num_servers

    def num_tuples(self, rel_key):
        if rel_key in self.sizes:
            return self.sizes[rel_key]
        return DEFAULT_CARDINALITY

    def partitioning(self, rel_key):
        if rel_key in self.partitionings:
            return RepresentationProperties(
                hash_partitioned=self.partitionings[rel_key])
        return RepresentationProperties()

    def get_scheme(self, rel_key):
        raise NotImplementedError()

    def get_function(self, funcName):
        """Return UDF with name = funcName"""
        if funcName in self.functions:
            return self.functions[funcName]


class FromFileCatalog(Catalog):

    """ Catalog that is created from a python file.
    Format of file is a dictionary of schemas.

    {'relation1' : [('a', 'LONG_TYPE'), ('b', 'STRING_TYPE')],
     'relation2' : [('y', 'STRING_TYPE'), ('z', 'DATETIME_TYPE')]}

     Or there can be an optional cardinality for any relation
    {'relation1' : ([('a', 'LONG_TYPE'), ('b', 'STRING_TYPE')], 10),
     'relation2' : [('y', 'STRING_TYPE'), ('z', 'DATETIME_TYPE')]}

     Or it can be a single relation, using filename as basename
     [('a', 'LONG_TYPE'), ('b', 'STRING_TYPE')]

     see raco.types for allowed types
    """

    def __init__(self, cat, fname):
        self.catalog = {}

        def error():
            assert False, """Unexpected catalog file format. \
                    See raco.catalog.FromFileCatalog"""

        if isinstance(cat, dict):
            def parse(v):
                if isinstance(v, tuple):
                    return v
                elif isinstance(v, list):
                    return v, DEFAULT_CARDINALITY
                else:
                    error()

            self.catalog = dict([(k, parse(v)) for k, v in cat.iteritems()])
        elif isinstance(cat, list):
            name = os.path.splitext(os.path.basename(fname))[0]
            self.catalog = {
                'public:adhoc:{0}'.format(name): (
                    cat,
                    DEFAULT_CARDINALITY,
                )}
        else:
            error()

    def get_scheme(self, rel_key):
        return Scheme(self.__get_catalog_entry__(rel_key)[0])

    def get_keys(self):
        return self.catalog.keys()

    @classmethod
    def print_cat(cls, ffc1, ffc2):
        cpy = ffc1.catalog.copy()
        for k, v in ffc2.catalog.iteritems():
            cpy[k] = v
        print cpy

    @classmethod
    def load_from_file(cls, path):
        with open(path) as fh:
            return cls(literal_eval(fh.read()), path)

    @classmethod
    def scheme_write_to_file(cls, path, new_rel_key, new_rel_schema,
                             append=True):
        new_schema_entry = literal_eval(new_rel_schema)
        col_names = new_schema_entry['columnNames']
        col_types = new_schema_entry['columnTypes']
        columns = zip(col_names, col_types)

        if os.path.isfile(path):
            if append:
                schema_read = open(path, 'r')
                s = schema_read.read()
                schema_read.close()

                schema_write = open(path, 'w')
                current_dict = literal_eval(s)
                current_dict[new_rel_key] = columns
                json.dump(current_dict, schema_write)
                schema_write.write("\n")
                schema_write.close()
            else:
                raise IOError("file {0} exists".format(path))
        else:
            with open(path, 'w+') as fh:
                d = {}
                d[new_rel_key] = columns
                json.dump(d, fh)
                fh.write("\n")
            fh.close()

    def get_num_servers(self):
        return 1

    def __get_catalog_entry__(self, rel_key):
        e = self.catalog.get(str(rel_key))
        if e is None:
            raise Exception(
                "relation {r} not found in catalog".format(r=rel_key))

        return e

    def num_tuples(self, rel_key):
        return self.__get_catalog_entry__(rel_key)[1]

    def partitioning(self, rel_key):
        # TODO allow specifying an optional list of attributes
        return RepresentationProperties()



import abc
from abc import abstractmethod
from raco.expression.function import UnaryFunction, \
    BinaryFunction, NaryFunction


class ExpressionVisitor(object):
    # TODO: make this more complete for kinds of expressions

    __metaclass__ = abc.ABCMeta

    def visit(self, expr):
        # use expr to dispatch to appropriate visit_* method

        # special case names for things we expect are not useful to
        # differentiate further
        # E.g., instead of visit_SQRT, do visit_UnaryFunction
        if isinstance(expr, UnaryFunction):
            typename = UnaryFunction.__name__
        elif isinstance(expr, NaryFunction):
            typename = NaryFunction.__name__
        else:
            # get most names from the expression's type
            typename = type(expr).__name__

        dispatchTo = getattr(self, "visit_%s" % (typename,))
        return dispatchTo(expr)

    @abc.abstractmethod
    def visit_NOT(self, unaryExpr):
        return

    @abc.abstractmethod
    def visit_AND(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_OR(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_EQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_NEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_GT(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_LT(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_GTEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_LTEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_NamedAttributeRef(self, named):
        return

    @abc.abstractmethod
    def visit_UnnamedAttributeRef(self, unnamed):
        return

    @abstractmethod
    def visit_NamedStateAttributeRef(self, attr):
        return

    @abc.abstractmethod
    def visit_StringLiteral(self, stringLiteral):
        return

    @abc.abstractmethod
    def visit_NumericLiteral(self, numericLiteral):
        return

    @abc.abstractmethod
    def visit_DIVIDE(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_PLUS(self, binaryExpr):
        return

    @abstractmethod
    def visit_MINUS(self, binaryExpr):
        return

    @abstractmethod
    def visit_IDIVIDE(self, binaryExpr):
        return

    @abstractmethod
    def visit_MOD(self, binaryExpr):
        return

    @abstractmethod
    def visit_TIMES(self, binaryExpr):
        return

    @abstractmethod
    def visit_NEG(self, unaryExpr):
        return

    @abstractmethod
    def visit_Case(self, caseExpr):
        return

    @abstractmethod
    def visit_UnaryFunction(self, expr):
        return

    @abstractmethod
    def visit_BinaryFunction(self, expr):
        return

    @abstractmethod
    def visit_NaryFunction(self, expr):
        return

    @abstractmethod
    def visit_CAST(self, expr):
        return

    @abstractmethod
    def visit_LIKE(self, binaryExpr):
        return


class SimpleExpressionVisitor(ExpressionVisitor):

    @abstractmethod
    def visit_unary(self, unaryexpr):
        pass

    @abstractmethod
    def visit_binary(self, binaryexpr):
        pass

    @abstractmethod
    def visit_zeroary(self, zeroaryexpr):
        pass

    @abstractmethod
    def visit_literal(self, literalexpr):
        pass

    @abstractmethod
    def visit_nary(self, naryexpr):
        pass

    @abstractmethod
    def visit_attr(self, attr):
        pass

    def visit_NOT(self, unaryExpr):
        self.visit_unary(unaryExpr)

    def visit_AND(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_OR(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_EQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_GT(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_LT(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_GTEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_LTEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NamedAttributeRef(self, named):
        self.visit_attr(named)

    def visit_UnnamedAttributeRef(self, unnamed):
        self.visit_attr(unnamed)

    def visit_NamedStateAttributeRef(self, attr):
        self.visit_attr(attr)

    def visit_StringLiteral(self, stringLiteral):
        self.visit_literal(stringLiteral)

    def visit_NumericLiteral(self, numericLiteral):
        self.visit_literal(numericLiteral)

    def visit_DIVIDE(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_PLUS(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_MINUS(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_IDIVIDE(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_MOD(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_TIMES(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NEG(self, unaryExpr):
        self.visit_unary(unaryExpr)

    def visit_LIKE(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_UnaryFunction(self, expr):
        self.visit_unary(expr)

    def visit_BinaryFunction(self, expr):
        self.visit_binary(expr)

    def visit_NaryFunction(self, expr):
        self.visit_nary(expr)

    def visit_CAST(self, expr):
        self.visit_unary(expr)
